#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
This file contains the main code for running CT and WD controlled models.
"""

import torch
import numpy as np
from projects.controllable_dialogue.tasks.build import build
from .stopwords import STOPWORDS
from .nidf import load_word2nidf
from .arora import SentenceEmbedder, load_arora


# Interrogative words, used to control question-asking via weighted decoding
# From https://en.wikipedia.org/wiki/Interrogative_word
QN_WORDS = ['who', 'what', 'where', 'why', 'when', 'how', 'which', 'whom', 'whose', '?']


# ========================================
# LOADING NIDF MEASURES
# ========================================


class NIDFFeats(object):
    """
    An object to hold a vector containing the NIDF values for all words in the
    vocabulary.

    The vector is contstructed when first needed.
    """

    def __init__(self):
        self.NIDF_FEATS = None  # will be vector length vocab_size containing NIDF vals

    def make_feat_vec(self, dict):
        """
        Construct the NIDF feature vector for the given dict.
        """
        print("Constructing NIDF feature vector...")
        self.NIDF_FEATS = torch.zeros((len(dict)))
        num_oovs = 0
        for idx in range(len(dict)):
            word = dict[idx]
            if word in word2nidf:
                # Leave emoji (these appear in Twitter dataset) as NIDF=0
                # (so we don't encourage emoji when we set WD weight high for NIDF)
                if word[0] == '@' and word[-1] == '@':
                    continue
                nidf = word2nidf[word]  # between 0 and 1
                self.NIDF_FEATS[idx] = nidf
            else:
                # print("WARNING: word %s has no NIDF; marking it as NIDF=0" % word)
                num_oovs += 1  # If we don't have NIDF for this word, set as 0
        print(
            'Done constructing NIDF feature vector; of %i words in dict there '
            'were %i words with unknown NIDF; they were marked as NIDF=0.'
            % (len(dict), num_oovs)
        )

    def get_feat_vec(self, dict):
        """
        Return the NIDF feature vector.

        If necessary, construct it first.
        """
        if self.NIDF_FEATS is None:
            self.make_feat_vec(dict)
        return self.NIDF_FEATS


word2nidf = None
nidf_feats = None
arora_data = None
sent_embedder = None


def initialize_control_information(opt, build_task=True):
    """
    Loads information from word2count.pkl, arora.pkl in data/controllable_dialogue, and
    uses it to initialize objects for computing NIDF and response-relatedness controls.

    By default (build_task=True) we will also build the controllable_dialogue task i.e.
    download data/controllable_dialogue if necessary.
    """
    global word2nidf, nidf_feats, arora_data, sent_embedder

    if word2nidf is not None:
        # already loaded, no need to do anything
        return

    if build_task:
        build(opt)

    print("Loading up controllable features...")
    word2nidf = load_word2nidf(opt)  # get word2nidf dict
    nidf_feats = NIDFFeats()  # init the NIDFFeats object
    # load info for arora sentence embeddings
    arora_data = load_arora(opt)
    sent_embedder = SentenceEmbedder(
        arora_data['word2prob'],
        arora_data['arora_a'],
        arora_data['glove_name'],
        arora_data['glove_dim'],
        arora_data['first_sv'],
        data_path=opt['datapath'],
    )


# ========================================
# UTIL
# ========================================


def flatten(list_of_lists):
    """
    Flatten a list of lists.
    """
    return [item for sublist in list_of_lists for item in sublist]


def intrep_frac(lst):
    """
    Returns the fraction of items in the list that are repeated.
    """
    if len(lst) == 0:
        return 0
    num_rep = 0
    for idx in range(len(lst)):
        if lst[idx] in lst[:idx]:
            num_rep += 1
    return num_rep / len(lst)


def extrep_frac(lst1, lst2):
    """
    Returns the fraction of items in lst1 that are in lst2.
    """
    if len(lst1) == 0:
        return 0
    num_rep = len([x for x in lst1 if x in lst2])
    return num_rep / len(lst1)


def get_ngrams(text, n):
    """
    Returns all ngrams that are in the text.

    Inputs:
        text: string
        n: int
    Returns:
        list of strings (each is a ngram)
    """
    tokens = text.split()
    return [
        " ".join(tokens[i : i + n]) for i in range(len(tokens) - (n - 1))
    ]  # list of str


def matching_ngram_completions(comparison_seq, hypothesis, n):
    """
    Return the list of words that if appended to hypothesis, would create a n-gram that
    already exists in comparison_seq. For efficiency, this function represents words as
    integers not strings.

    Inputs:
        comparison_seq: list of integers
        hypothesis: list of integers or None
        n: integer

    Output:
        bad_words: list of integers
    """
    if hypothesis is None or len(hypothesis) < n - 1 or len(comparison_seq) < n:
        return []
    hypothesis = [int(i) for i in hypothesis]  # cast to list of ints
    comparison_seq = [int(i) for i in comparison_seq]  # cast to list of ints
    n_minus_1_gram = hypothesis[-(n - 1) :]  # list of ints length n-1
    bad_words = [
        comparison_seq[i]
        for i in range(n - 1, len(comparison_seq))
        if comparison_seq[i - (n - 1) : i] == n_minus_1_gram
    ]  # list of ints
    return bad_words


# ========================================
# WEIGHTED DECODING FEATURE FUNCTIONS
# These functions compute the decoding features for weighted decoding.
#
# Given a conversational history and a hypothesis (i.e. a partially generated response),
# these functions update the weighted decoding feature vector (of length vocab_size) by
# adding the new decoding feature, multiplied by its corresponding weight.
#
# All these functions have the following inputs and outputs:
#
# Inputs:
#   dict: parlai DictionaryAgent
#   hypothesis: a list of integers. This is the partially generated response,
#     represented via word indices.
#   history: a ConvAI2History. This represents the conversation history.
#   wt: a float. This is the weight for the weighted decoding feature.
#   feat: a vector length vocab_size. This will ultimately contain the sum of all the
#     weighted decoding features.
# Output:
#   feat: a vector length vocab_size. This is the feature vector, now with the new
#     weighted decoding feature added (multiplied by wt).
# ========================================


def intrep_word_used_before(dict, hypothesis, history, wt, feat, remove_stopwords):
    """
    Weighted decoding feature function. See explanation above. This feature is 1 for
    words that have already appeared within the hypothesis, 0 otherwise.

    Additional inputs:
      remove_stopwords: bool. If True, stopwords are not included when identifying words
        that have already appeared.
    """
    if hypothesis is not None:
        if remove_stopwords:
            hypothesis = [idx for idx in hypothesis if dict[idx] not in STOPWORDS]
        if len(hypothesis) > 0:
            feat[hypothesis] += wt
    return feat


def intrep_ngram_used_before(dict, hypothesis, history, wt, feat, n):
    """
    Weighted decoding feature function. See explanation above. This feature is 1 for
    words that, if added to the hypothesis, will create a n-gram that has already
    appeared in the hypothesis; otherwise 0.

    Additional inputs:
      n: int, the size of the n-grams considered.
    """
    if hypothesis is not None:
        bad_words = matching_ngram_completions(hypothesis, hypothesis, n)
        if len(bad_words) > 0:
            feat[bad_words] += wt
    return feat


def extrep_word_used_before(
    dict, hypothesis, history, wt, feat, remove_stopwords, person
):
    """
    Weighted decoding feature function. See explanation above. This feature is 1 for
    words that have already been used earlier in the conversation; otherwise 0.

    Additional inputs:
      remove_stopwords: bool. If True, stopwords are not included when identifying words
        that have already appeared.
      person: If 'self', identify words that have already been used by self (bot).
        If 'partner', identify words that have already been used by partner (human).
    """
    if person == 'self':
        prev_utts = history.own_utts
    elif person == 'partner':
        prev_utts = history.partner_utts
    else:
        raise ValueError("person must be 'self' or 'partner', but it is: ", person)
    if len(prev_utts) == 0:
        return feat
    prev_words = [dict.txt2vec(utt) for utt in prev_utts]  # list of list of ints
    prev_words = list(set(flatten(prev_words)))  # list of ints, no duplicates
    if remove_stopwords:
        prev_words = [idx for idx in prev_words if dict[idx] not in STOPWORDS]
    if len(prev_words) > 0:
        feat[prev_words] += wt
    return feat


def extrep_ngram_used_before(dict, hypothesis, history, wt, feat, n, person):
    """
    Weighted decoding feature function. See explanation above. This feature is 1 for
    words that, if added to hypothesis, would create a n-gram that has already been used
    earlier in the conversation; otherwise 0.

    Additional inputs:
      n: int, the size of the n-grams considered.
      person: If 'self', identify n-grams that have already been used by self (bot).
        If 'partner', identify n-grams that have already been used by partner (human).
    """
    if person == 'self':
        prev_utts = history.own_utts
    elif person == 'partner':
        prev_utts = history.partner_utts
    else:
        raise ValueError("person must be 'self' or 'partner', but it is: ", person)
    if len(prev_utts) == 0:
        return feat
    if hypothesis is None:
        return feat
    prev_utts_wordidx = [dict.txt2vec(utt) for utt in prev_utts]  # list of list of ints
    bad_words = [
        matching_ngram_completions(prev_utt, hypothesis, n)
        for prev_utt in prev_utts_wordidx
    ]  # list of list of ints
    bad_words = list(set(flatten(bad_words)))  # list of ints, no duplicates
    if len(bad_words) > 0:
        feat[bad_words] += wt
    return feat


def nidf(dict, hypothesis, history, wt, feat):
    """
    Weighted decoding feature function.

    See explanation above. This feature is equal to the NIDF (normalized inverse
    document frequency) score for each word. The score is always between 0 and 1.
    """
    feat += wt * nidf_feats.get_feat_vec(dict)
    return feat


def qn_words(dict, hypothesis, history, wt, feat):
    """
    Weighted decoding feature function.

    See explanation above. This feature is 1 for 'interrogative words', 0 otherwise.
    """
    qn_indices = [dict[w] for w in QN_WORDS]
    feat[qn_indices] += wt
    return feat


def lastutt_sim_arora_word(dict, hypothesis, history, wt, feat):
    """
    Weighted decoding feature function.

    See explanation above. Given a word w, this feature is equal to cos_sim(word_emb(w),
    sent_emb(l)) the cosine similarity between the GloVe vector for word w, and the
    Arora-style sentence embedding for the partner's last utterance l.
    """
    partner_utts = history.partner_utts
    if len(partner_utts) == 0:  # if bot goes first then do nothing
        return feat
    last_utt = partner_utts[-1]  # string
    if last_utt.strip().lower() == "__silence__":  # if bot goes first then do nothing
        return feat

    # Get last_utt_emb, which is a tensor shape (glove_dim)
    last_utt_emb = sent_embedder.embed_sent(dict.tokenize(last_utt))
    if last_utt_emb is None:
        return feat

    # Get cosine similarities, which is a tensor shape (vocab_size)
    sims = sent_embedder.get_word_sims(last_utt, last_utt_emb, dict)

    feat += wt * sims
    return feat


# In this dictionary, the keys are the names of the WD features, and the values are
# functions with inputs (dict, hypothesis, history, wt, feat), that update the feature
# vector feat.
WDFEATURE2UPDATEFN = {
    # Use to reduce repeated words within an utterance. Not used in paper.
    "intrep_word": (
        lambda x: intrep_word_used_before(
            x[0], x[1], x[2], x[3], x[4], remove_stopwords=False
        )
    ),
    # Use to reduce repeated non-stopwords within an utterance. intrep_unigram in paper.
    "intrep_nonstopword": (
        lambda x: intrep_word_used_before(
            x[0], x[1], x[2], x[3], x[4], remove_stopwords=True
        )
    ),
    # Use to reduce repeated 2-grams within an utterance. intrep_bigram in paper.
    "intrep_2gram": (
        lambda x: intrep_ngram_used_before(x[0], x[1], x[2], x[3], x[4], n=2)
    ),
    # Use to reduce repeated 3-grams within an utterance. Not used in paper.
    "intrep_3gram": (
        lambda x: intrep_ngram_used_before(x[0], x[1], x[2], x[3], x[4], n=3)
    ),
    # Use to reduce repeating words already used in previous bot utterances.
    # Not used in paper.
    "extrep_word": (
        lambda x: extrep_word_used_before(
            x[0], x[1], x[2], x[3], x[4], remove_stopwords=False, person='self'
        )
    ),
    # Use to reduce repeating non-stopwords already used in previous bot utterances.
    # extrep_unigram in paper.
    "extrep_nonstopword": (
        lambda x: extrep_word_used_before(
            x[0], x[1], x[2], x[3], x[4], remove_stopwords=True, person='self'
        )
    ),
    # Use to reduce repeating 2-grams already used in previous bot utterances.
    # extrep_bigram in paper.
    "extrep_2gram": (
        lambda x: extrep_ngram_used_before(
            x[0], x[1], x[2], x[3], x[4], n=2, person='self'
        )
    ),
    # Use to reduce repeating 3-grams already used in previous bot utterances.
    # Not used in paper.
    "extrep_3gram": (
        lambda x: extrep_ngram_used_before(
            x[0], x[1], x[2], x[3], x[4], n=3, person='self'
        )
    ),
    # Use to reduce repeating words already used in previous partner utterances.
    # Not used in paper.
    "partnerrep_word": (
        lambda x: extrep_word_used_before(
            x[0], x[1], x[2], x[3], x[4], remove_stopwords=False, person='partner'
        )
    ),
    # Use to reduce repeating non-stopwords already used in previous partner utterances.
    # Not used in paper.
    "partnerrep_nonstopword": (
        lambda x: extrep_word_used_before(
            x[0], x[1], x[2], x[3], x[4], remove_stopwords=True, person='partner'
        )
    ),
    # Use to reduce repeating 2-grams already used in previous partner utterances.
    # partnerrep_bigram in paper.
    "partnerrep_2gram": (
        lambda x: extrep_ngram_used_before(
            x[0], x[1], x[2], x[3], x[4], n=2, person='partner'
        )
    ),
    # Use to reduce repeating 3-grams already used in previous partner utterances.
    # Not used in paper.
    "partnerrep_3gram": (
        lambda x: extrep_ngram_used_before(
            x[0], x[1], x[2], x[3], x[4], n=3, person='partner'
        )
    ),
    # Use to increase/decrease the probability of high-specificity (i.e. rare) words.
    # This is the NIDF(w) weighted decoding feature mentioned in the paper.
    "nidf": (lambda x: nidf(x[0], x[1], x[2], x[3], x[4])),
    # Use to increase/decrease the probability of interrogative (i.e. question) words.
    # This is the is_qn_word(w) WD feature mentioned in the paper.
    "question": (lambda x: qn_words(x[0], x[1], x[2], x[3], x[4])),
    # Use to increase/decrease the probability of words with high response-relatedness
    # (i.e. similarity to the partner's last utterance).
    # This is the resp_rel(w) WD feature mentioned in the paper.
    "lastuttsim": (lambda x: lastutt_sim_arora_word(x[0], x[1], x[2], x[3], x[4])),
}


def get_wd_features(dict, hypothesis, history, wd_features, wd_weights):
    """
    Given a conversational history and a hypothesis (i.e. partially generated response),
    compute the Weighted Decoding features for all words in the vocabulary.

    Inputs:
        dict: parlai DictionaryAgent
        hypothesis: list of ints or None
        history: a ConvAI2History. This represents the conversation history.
        wd_features: list of strings; the names of the WD features we want to use
        wd_weights: list of floats; the weights corresponding to the WD features.
    Returns:
        wd_feat_vec: tensor shape (vocab_size), containing weighted sum of the feature
          functions, for each candidate continuation word
    """
    wd_feat_vec = torch.zeros((len(dict)))
    for f, w in zip(wd_features, wd_weights):
        wd_feat_vec = WDFEATURE2UPDATEFN[f]((dict, hypothesis, history, w, wd_feat_vec))
    return wd_feat_vec


# ========================================
# SENTENCE-LEVEL ATTRIBUTE FUNCTIONS
# Given an input utterance, these functions compute the value of the controllable
# attribute at the sentence level (more precisely, at the utterance level).
#
# All these functions have the following inputs and outputs:
#
# Inputs:
#   utt: a string, tokenized and lowercase
#   history: a ConvAI2History. This represents the conversation history.
# Output:
#   score: float. the value of the controllable attribute for utt.
# ========================================


def intrep_repeated_word_frac(utt, history, remove_stopwords):
    """
    Sentence-level attribute function.

    See explanation above.
    Returns the fraction of words in utt that are repeated.
    Additional inputs:
      remove_stopwords: bool. If True, stopwords are removed before counting repetition.
    """
    assert utt.strip() != ""
    tokens = utt.split()
    if remove_stopwords:
        tokens = [t for t in tokens if t not in STOPWORDS]
    return intrep_frac(tokens)


def intrep_repeated_ngram_frac(utt, history, n):
    """
    Sentence-level attribute function.

    See explanation above.
    Returns the fraction of n-grams in utt that are repeated.
    Additional inputs:
      n: int, the size of the n-grams considered.
    """
    assert utt.strip() != ""
    ngrams = get_ngrams(utt, n)
    return intrep_frac(ngrams)


def extrep_repeated_word_frac(utt, history, remove_stopwords, person):
    """
    Sentence-level attribute function.

    See explanation above.
    Returns the fraction of words in utt that already appeared in a previous utterance.
    Additional inputs:
      remove_stopwords: bool. If True, stopwords are removed from utt before counting
        repetition.
      person: If 'self', identify words that have already been used by self (bot).
        If 'partner', identify words that have already been used by partner (human).
    """
    assert utt.strip() != ""
    if person == 'self':
        prev_utts = history.own_utts  # should already be tokenized
    elif person == 'partner':
        prev_utts = history.partner_utts  # should already be tokenized
    else:
        raise ValueError("person must be 'self' or 'partner', but it is: ", person)
    if len(prev_utts) == 0:
        return 0
    tokens = utt.split()  # list of strings
    if remove_stopwords:  # remove stopwords from utt
        tokens = [t for t in tokens if t not in STOPWORDS]
    prev_words = [s.split() for s in prev_utts]  # list of list of ints
    prev_words = list(set(flatten(prev_words)))  # list of ints, no duplicates
    return extrep_frac(tokens, prev_words)


def extrep_repeated_ngram_frac(utt, history, n, person):
    """
    Sentence-level attribute function.

    See explanation above.
    Returns fraction of n-grams in utt that already appeared in a previous utterance.
    Additional inputs:
      n: int, the size of the n-grams considered.
      person: If 'self', identify n-grams that have already been used by self (bot).
        If 'partner', identify n-grams that have already been used by partner (human).
    """
    assert utt.strip() != ""
    if person == 'self':
        prev_utts = history.own_utts  # should already be tokenized
    elif person == 'partner':
        prev_utts = history.partner_utts  # should already be tokenized
    else:
        raise ValueError("person must be 'self' or 'partner', but it is: ", person)
    if len(prev_utts) == 0:
        return 0
    utt_ngrams = get_ngrams(utt, n)
    prev_ngrams = [get_ngrams(prev, n) for prev in prev_utts]  # list of list of strings
    prev_ngrams = list(set(flatten(prev_ngrams)))  # list of strings, no duplicates
    return extrep_frac(utt_ngrams, prev_ngrams)


def avg_nidf(utt, history):
    """
    Sentence-level attribute function.

    See explanation above. Returns the mean NIDF of the words in utt.
    """
    words = utt.split()
    problem_words = [w for w in words if w not in word2nidf]
    ok_words = [w for w in words if w in word2nidf]
    if len(ok_words) == 0:
        print(
            "WARNING: For all the words in the utterance '%s', we do not have the "
            "NIDF score. Marking as avg_nidf=1." % utt
        )
        return 1  # rarest possible sentence
    nidfs = [word2nidf[w] for w in ok_words]
    avg_nidf = sum(nidfs) / len(nidfs)
    if len(problem_words) > 0:
        print(
            "WARNING: When calculating avg_nidf for the utterance '%s', we don't "
            "know NIDF for the following words: %s" % (utt, str(problem_words))
        )
    assert avg_nidf >= 0 and avg_nidf <= 1
    return avg_nidf


def contains_qmark(utt, history):
    """
    Sentence-level attribute function.

    See explanation above. Returns 1 if utt contains a question mark, otherwise 0.
    """
    return int("?" in utt)


def lastutt_sim_arora_sent(utt, history):
    """
    Sentence-level attribute function. See explanation above.

    Returns
      cos_sim(sent_emb(last_utt), sent_emb(utt))
    the cosine similarity of the Arora-style sentence embeddings for the current
    response (utt) and the partner's last utterance (last_utt, which is in history).

    - If there is no last_utt (i.e. utt is the first utterance of the conversation),
      returns None.
    - If one or both of utt and last_utt are all-OOV; thus we can't compute sentence
      embeddings, return the string 'oov'.
    """
    partner_utts = history.partner_utts
    if len(partner_utts) == 0:
        # print('WARNING: returning lastuttsim = None because bot goes first')
        return None
    last_utt = partner_utts[-1]  # string
    if "__SILENCE__" in last_utt:
        assert last_utt.strip() == "__SILENCE__"
        # print('WARNING: returning lastuttsim = None because bot goes first')
        return None

    # Get sentence embeddings. Here we're naively splitting last_utt and utt; this is
    # fine given that we assume both utt and history are lowercase and tokenized.
    # Both last_utt_emb and response_emb are tensors length glove_dim (or None)
    last_utt_emb = sent_embedder.embed_sent(last_utt.split())
    response_emb = sent_embedder.embed_sent(utt.split())
    if last_utt_emb is None or response_emb is None:
        return 'oov'

    sim = torch.nn.functional.cosine_similarity(last_utt_emb, response_emb, dim=0)
    return sim.item()


def wordlist_frac(utt, history, word_list):
    """
    Sentence-level attribute function.

    See explanation above.
    Returns the fraction of words in utt that are in word_list.
    Additional inputs:
      word_list: list of strings.
    """
    words = utt.split()
    num_in_list = len([w for w in words if w in word_list])
    return num_in_list / len(words)


# In this dict, the keys are the names of the sentence-level attributes, and the values
# are functions with input (utt, history), returning the attribute value measured on utt
ATTR2SENTSCOREFN = {
    # Proportion of words in utt that appear earlier in utt
    "intrep_word": (
        lambda x: intrep_repeated_word_frac(x[0], x[1], remove_stopwords=False)
    ),
    # Proportion of non-stopwords in utt that appear earlier in utt
    "intrep_nonstopword": (
        lambda x: intrep_repeated_word_frac(x[0], x[1], remove_stopwords=True)
    ),
    # Proportion of 2-grams in utt that appear earlier in utt
    "intrep_2gram": (lambda x: intrep_repeated_ngram_frac(x[0], x[1], n=2)),
    # Proportion of 3-grams in utt that appear earlier in utt
    "intrep_3gram": (lambda x: intrep_repeated_ngram_frac(x[0], x[1], n=3)),
    # Proportion of words in utt that appeared in a previous bot utterance
    "extrep_word": (
        lambda x: extrep_repeated_word_frac(
            x[0], x[1], remove_stopwords=False, person='self'
        )
    ),
    # Proportion of non-stopwords in utt that appeared in a previous bot utterance
    "extrep_nonstopword": (
        lambda x: extrep_repeated_word_frac(
            x[0], x[1], remove_stopwords=True, person='self'
        )
    ),
    # Proportion of 2-grams in utt that appeared in a previous bot utterance
    "extrep_2gram": (
        lambda x: extrep_repeated_ngram_frac(x[0], x[1], n=2, person='self')
    ),
    # Proportion of 3-grams in utt that appeared in a previous bot utterance
    "extrep_3gram": (
        lambda x: extrep_repeated_ngram_frac(x[0], x[1], n=3, person='self')
    ),
    # Proportion of words in utt that appeared in a previous partner utterance
    "partnerrep_word": (
        lambda x: extrep_repeated_word_frac(
            x[0], x[1], remove_stopwords=False, person='partner'
        )
    ),
    # Proportion of non-stopwords in utt that appeared in a previous partner utterance
    "partnerrep_nonstopword": (
        lambda x: extrep_repeated_word_frac(
            x[0], x[1], remove_stopwords=True, person='partner'
        )
    ),
    # Proportion of 2-grams in utt that appeared in a previous partner utterance
    "partnerrep_2gram": (
        lambda x: extrep_repeated_ngram_frac(x[0], x[1], n=2, person='partner')
    ),
    # Proportion of 3-grams in utt that appeared in a previous partner utterance
    "partnerrep_3gram": (
        lambda x: extrep_repeated_ngram_frac(x[0], x[1], n=3, person='partner')
    ),
    # Mean NIDF score of the words in utt
    "avg_nidf": (lambda x: avg_nidf(x[0], x[1])),
    # 1 if utt contains '?', 0 otherwise
    "question": (lambda x: contains_qmark(x[0], x[1])),
    # Proportion of words in utt that are interrogative words
    "qn_words": (lambda x: wordlist_frac(x[0], x[1], word_list=QN_WORDS)),
    # Cosine similarity of utt to partner's last utterance
    "lastuttsim": (lambda x: lastutt_sim_arora_sent(x[0], x[1])),
}


def eval_attr(utt, history, attr):
    """
    Given a conversational history and an utterance, compute the requested sentence-
    level attribute for utt.

    Inputs:
        utt: string. The utterance, tokenized and lowercase
        history: a ConvAI2History. This represents the conversation history.
        attr: string. The name of the sentence-level attribute.
    Returns:
        value: float. The value of the attribute for utt.
    """
    # Check everything is lowercased already
    assert utt == utt.lower()
    for thing in [history.persona_lines, history.partner_utts, history.own_utts]:
        for line in thing:
            if line != "__SILENCE__":
                assert line == line.lower()

    # Eval attribute
    return ATTR2SENTSCOREFN[attr]((utt, history))


# ========================================
# GETTING CONTROL VARIABLE BUCKETS
# For Conditional Training (CT) models, the code in this section allows us to determine
# what bucket a given control variable value should go into.
# ========================================


def get_qn_bucket_probs():
    """
    Assuming we have 11 CT question buckets (0 to 10), compute P(bucket|question=1) and
    P(bucket|question=0); this is needed so we can probabilistically assign incoming
    training examples to buckets.

    Returns:
      prob_bucket_given_qn: list of floats length 11; P(bucket|question=1)
      prob_bucket_given_notqn: list of floats length 11; P(bucket|question=0)
    """
    prob_qn = 41101 / 131438  # P(question=1), computed across ConvAI2 dataset. ~31%

    # Compute P(bucket), i.e. the total sizes of the buckets.
    # This is done by assuming that buckets 1 to 10 are equal in size, but bucket 0
    # is larger because we have more non-questions than questions. Therefore:
    # P(question=1) = P(bucket=1, question=1) + ... + P(bucket=10, question=1)
    #               = prob_bucket_n * 0.1 + ... + prob_bucket_n * 1
    #               = 5.5 * prob_bucket_n
    # Thus we can derive the value for prob_bucket_n and prob_bucket_0:
    prob_bucket_n = prob_qn / 5.5  # P(bucket=n) for n=1,...,10
    prob_bucket_0 = 1 - 10 * prob_bucket_n  # P(bucket=0)
    prob_bucket = [prob_bucket_0] + [prob_bucket_n] * 10  # list length 11, P(bucket)

    # Compute P(bucket|qn=1) and P(bucket|qn=0) using Bayes Rule:
    # P(bucket|qn=1) = P(bucket) * P(qn=1|bucket) / P(qn=1)
    # P(bucket|qn=0) = P(bucket) * P(qn=0|bucket) / P(qn=0)
    prob_bucket_given_qn = [pb * (i / 10) / prob_qn for i, pb in enumerate(prob_bucket)]
    prob_bucket_given_notqn = [
        pb * ((10 - i) / 10) / (1 - prob_qn) for i, pb in enumerate(prob_bucket)
    ]

    return prob_bucket_given_qn, prob_bucket_given_notqn


PROB_BUCKET_GIVEN_QN, PROB_BUCKET_GIVEN_NOTQN = get_qn_bucket_probs()


def bucket_question(ex, ctrl, num_buckets):
    """
    Given an example (where the target response may or may not be a question) and its
    history, probabilistically determine what question-asking CT bucket to use.

    Inputs:
      ex: message dictionary containing a bool field 'question'
      ctrl: string. The name of the CT control. Should be 'question'.
      num_buckets: int. The number of question-asking CT buckets. Assumed to be 11.
    Returns:
      out: int. bucket number.
    """
    assert num_buckets == 11
    is_qn = int(ex['question'])
    assert is_qn in [0, 1]
    is_qn = bool(is_qn)
    if is_qn:  # Sample from P(bucket|qn=1)
        out = np.random.choice(range(num_buckets), 1, p=PROB_BUCKET_GIVEN_QN)
    else:  # Sample from P(bucket|qn=0)
        out = np.random.choice(range(num_buckets), 1, p=PROB_BUCKET_GIVEN_NOTQN)
    out = int(out[0])
    return out


def sort_into_bucket(val, bucket_lbs):
    """
    Returns the highest bucket such that val >= lower bound for that bucket.

    Inputs:
      val: float. The value to be sorted into a bucket.
      bucket_lbs: list of floats, sorted ascending.

    Returns:
      bucket_id: int in range(num_buckets); the bucket that val belongs to.
    """
    num_buckets = len(bucket_lbs)
    for bucket_id in range(num_buckets - 1, -1, -1):  # iterate descending
        lb = bucket_lbs[bucket_id]
        if val >= lb:
            return bucket_id
    raise ValueError('val %f is not >= any of the lower bounds: %s' % (val, bucket_lbs))


def bucket_contvar(ex, ctrl, num_buckets):
    """
    Given ex, which contains a continuous value for a particular control variable,
    return the bucketed version of that control value.

    Inputs:
      ex: message dictionary. Assume it has key ctrl, mapping to the value.
      ctrl: string. The name of the CT control.
      num_buckets: int. The number of buckets for this control variable.
    """
    if ctrl not in ex.keys():
        raise ValueError(
            "Control %s not found in example. Available keys in "
            "this example: %s" % (ctrl, ', '.join(ex.keys()))
        )

    # Get the control variable value
    ctrl_val = ex[ctrl]  # string. the value of the control variable for this example
    if ctrl == 'avg_nidf':
        ctrl_val = float(ctrl_val)
        assert ctrl_val >= 0
        assert ctrl_val <= 1
    elif ctrl == 'lastuttsim':
        if ctrl_val == 'None':  # bot goes first in conversation
            assert num_buckets == 11
            return 10  # The last bucket is for when the bot goes first
        else:
            ctrl_val = float(ctrl_val)
            assert ctrl_val >= -1
            assert ctrl_val <= 1
    else:
        raise ValueError('Unexpected CT ctrl: %s' % ctrl)

    # Get the bucket lowerbounds
    bucket_lbs = CONTROL2BUCKETLBS[(ctrl, num_buckets)]  # lst len num_buckets of floats
    if ctrl == 'lastuttsim':
        # The 'bot goes first' bucket 10 has no lower bound
        assert len(bucket_lbs) == num_buckets - 1
    else:
        assert len(bucket_lbs) == num_buckets

    # Determine the correct bucket and return the bucket id
    return sort_into_bucket(ctrl_val, bucket_lbs)


# The default embedding size for CT control variable embeddings
CONTROL2DEFAULTEMBSIZE = {'question': 10, 'avg_nidf': 10, 'lastuttsim': 10}

# The default number of buckets for CT control variables
CONTROL2DEFAULTNUMBUCKETS = {
    'question': 11,
    'avg_nidf': 10,
    'lastuttsim': 11,  # 11th bucket is for when the bot goes first in the conversation
}

# This dictionary maps from the name of a CT control variable, to a function that
# takes (ex, ctrl, num_buckets) as input, and returns the correct bucket_id
# for that control and this example.
CONTROL2BUCKETINGFN = {
    'question': bucket_question,
    'avg_nidf': bucket_contvar,
    'lastuttsim': bucket_contvar,
}

# Bucket lowerbounds. These are produced using the get_bucket_lowerbounds.py script.
AVG_NIDF_10BUCKET_LBS = [
    0.0,
    0.1598414705378728,
    0.17498045049881217,
    0.18658836637678175,
    0.19671787445075514,
    0.2070643776875113,
    0.2182630256396894,
    0.23053753067016441,
    0.24624559431359425,
    0.2707252238670671,
]
LASTUTTSIM_10BUCKET_LBS = [
    -0.3870984613895416,
    -0.08026778697967529,
    -0.025567850098013878,
    0.019155802205204964,
    0.06262511014938354,
    0.10953287780284882,
    0.16335178911685944,
    0.2319537252187729,
    0.3283223509788513,
    0.4921867549419403,
]


# This dictionary maps from (CT control variable name, num_buckets) to the lower bounds
CONTROL2BUCKETLBS = {
    ('avg_nidf', 10): AVG_NIDF_10BUCKET_LBS,
    # Note: For lastuttsim, the 11th bucket is for when the bot goes first; it doesn't
    # have a LB but it does have an embedding.
    ('lastuttsim', 11): LASTUTTSIM_10BUCKET_LBS,
}


def get_ctrl_vec(exs, history, control_settings):
    """
    Given a batch of examples with given history, return the bucketed CT control values.
    This is used both when training and evaluating CT systems.

    Inputs:
      exs: list length batch_size of message dictionaries. Each dictionary contains
        a 'text' field, and a field for each CT control we're using, along with the
        value of the CT control variable.
      history: list length batch_size of ConvAI2History objects. These represent the
        conversation history.
      control_settings: dictionary containing info about CT controls.
        See ControllableSeq2seqAgent.control_settings.

    Returns:
      ctrl_vec: torch Tensor shape (batch_size, num_controls), with the bucketed values
        for the CT controls we're using. If there's no CT controls, return None.
    """
    if len(control_settings) == 0:
        return None

    # ctrl_vec is shape (bsz, num_controls) filled with -1's
    ctrl_vec = -torch.ones((len(exs), len(control_settings))).long()

    for batch_idx, ex in enumerate(exs):
        for ctrl, ctrl_info in control_settings.items():
            set_val = ctrl_info['set_value']  # is either int or None
            if set_val is not None:  # if we're using some preset bucket for this ctrl
                bucket = set_val  # override with set_val, an int
            else:  # bucket the control val given in ex
                if ctrl not in ex:
                    raise ValueError(
                        "The CT control '%s' is not present as a key in "
                        "the message dictionary:\n%s\nIf training a CT "
                        "model, perhaps your training data is missing the "
                        "annotations. If talking interactively, perhaps "
                        "you forgot to set --set-controls." % (ctrl, str(ex))
                    )
                num_buckets = ctrl_info['num_buckets']
                bucketing_fn = CONTROL2BUCKETINGFN[ctrl]  # bucketing fn for this ctrl
                bucket = bucketing_fn(ex, ctrl, num_buckets)  # int

            # If we have multiple CT controls, ctrl_idx tells us which order they go in
            ctrl_idx = ctrl_info['idx']  # int
            ctrl_vec[batch_idx, ctrl_idx] = bucket
    return ctrl_vec



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
9.4B parameter Blender model: please see <parl.ai/project/blender>.
"""

from .build import build

VERSION = 'v1.0'


def download(datapath):
    build(datapath, 'BST9B.tgz', model_type='blender_9B', version=VERSION)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


"""
Main launch script for single-host, multi-GPU evaluation.

This is a drop-in replacement for [eval_model]. This script will launch N
subprocess, each which runs the full eval loop independently.

Uses torch.nn.parallel.DistributedDataParallel for its main uses. Agents
must specifically implement the wrapper of DistributedDataParallel, but
all TorchRankerAgents and TorchGeneratorAgents support this.

## Examples

```shell
parlai multiprocessing_eval --model-file "zoo:tutorial_transformer_generator/model" --batchsize 16 --task convai2
```
"""

import torch
import random
import os
import signal
import parlai.utils.distributed as distributed_utils
import parlai.scripts.eval_model as eval_model
from parlai.core.script import ParlaiScript, register_script


def multiprocess_eval(
    rank, opt, port=61337, rank_offset=0, gpu=None, hostname='localhost'
):
    """
    Run a multiprocessing evaluation.

    Invoked by launch_and_eval, not instantiated directly.
    """
    init_method = f'tcp://{hostname}:{port}'
    with distributed_utils.distributed_context(
        rank, opt, rank_offset, gpu, init_method=init_method
    ) as opt:
        opt['multiprocessing'] = True
        return eval_model.eval_model(opt)


def launch_and_eval(opt, port):
    """
    Perform a fork() to many processes.
    """
    # Launch multiple subprocesses
    spawncontext = torch.multiprocessing.spawn(
        multiprocess_eval,
        # need to give rank offset as 1 to cover the fact that the main
        # process is rank 0, but that spawn() doesn't let you control rank
        (opt, port, 1),
        nprocs=opt['distributed_world_size'] - 1,  # main proc will also run loop
        join=False,
    )

    try:
        retval = multiprocess_eval(0, opt, port)
        spawncontext.join()
        return retval
    except KeyboardInterrupt:
        # tell the subprocesses to stop too
        for p in spawncontext.processes:
            if p.is_alive():
                os.kill(p.pid, signal.SIGINT)
        raise


def setup_args():
    parser = eval_model.setup_args()
    parser.add_distributed_training_args()
    parser.set_defaults(distributed_world_size=torch.cuda.device_count())
    return parser


@register_script("multiprocessing_eval", aliases=["mp_eval"], hidden=True)
class MultiProcessEval(ParlaiScript):
    @classmethod
    def setup_args(cls):
        return setup_args()

    def run(self):
        port = random.randint(32000, 48000)
        return launch_and_eval(self.opt, port)


if __name__ == '__main__':
    MultiProcessEval.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# Download and build the data if it does not exist.

import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile

RESOURCES = [
    DownloadableFile(
        'https://dada.cs.washington.edu/qasrl/data/wiki1.train.qa',
        'wiki1.train.qa',
        'f927417e94e67b7ae17e33dd882989a5556d7ff37376f8bf5c78ece7d17a6c64',
        zipped=False,
    ),
    DownloadableFile(
        'https://dada.cs.washington.edu/qasrl/data/wiki1.dev.qa',
        'wiki1.dev.qa',
        'caa94beaaf22304422cdc1a2fd8732b1a47401c9555a81e1f4da81e0a7557a8b',
        zipped=False,
    ),
    DownloadableFile(
        'https://dada.cs.washington.edu/qasrl/data/wiki1.test.qa',
        'wiki1.test.qa',
        'b43a998344fbd520955fb8f0f7b3691ace363daa8628552cf5cf5c8d84df6cca',
        zipped=False,
    ),
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'QA-SRL')
    version = None

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
File for miscellaneous utility functions and constants.
"""

from collections import deque, OrderedDict
from typing import Union, Optional, Set, Any, Dict, List, Tuple
from datetime import timedelta
import functools
import math
import time
import re
import shutil
import json
import os

from parlai.core.message import Message
from parlai.utils.strings import colorize
from parlai.utils.io import PathManager
import parlai.utils.logging as logging

try:
    import torch

    __TORCH_AVAILABLE = True
except ImportError:
    # silence the error, we'll have other problems later if it's super necessary
    __TORCH_AVAILABLE = False


SPECIAL_FORMATED_DISPLAY_MESSAGE_FIELDS = {
    'episode_done',
    'id',
    'image',
    'text',
    'labels',
    'eval_labels',
    'label_candidates',
    'text_candidates',
    'reward',
    'token_losses',
    'metrics',
}

MUST_SHOW_MESSAGE_FIELDS = {'image', 'text', 'labels', 'eval_labels', 'reward'}


def maintain_dialog_history(
    history,
    observation,
    reply='',
    historyLength=1,
    useReplies='label_else_model',
    dict=None,
    useStartEndIndices=True,
    splitSentences=False,
):
    """
    Keep track of dialog history, up to a truncation length.

    Either includes replies from the labels, model, or not all using param
    'replies'.

    DEPRECATED. USE PARLAI.CORE.TORCH_AGENT INSTEAD.
    """

    def parse(txt, splitSentences):
        if dict is not None:
            if splitSentences:
                vec = [dict.txt2vec(t) for t in txt.split('\n')]
            else:
                vec = dict.txt2vec(txt)
            return vec
        else:
            return [txt]

    if 'dialog' not in history:
        history['dialog'] = deque(maxlen=historyLength)
        history['episode_done'] = False
        history['labels'] = []

    if history['episode_done']:
        history['dialog'].clear()
        history['labels'] = []
        useReplies = 'none'
        history['episode_done'] = False

    if useReplies != 'none':
        if useReplies == 'model' or (
            useReplies == 'label_else_model' and len(history['labels']) == 0
        ):
            if reply:
                if useStartEndIndices:
                    reply = dict.start_token + ' ' + reply
                history['dialog'].extend(parse(reply, splitSentences))
        elif len(history['labels']) > 0:
            r = history['labels'][0]
            history['dialog'].extend(parse(r, splitSentences))

    obs = observation
    if 'text' in obs:
        if useStartEndIndices:
            obs['text'] = dict.end_token + ' ' + obs['text']
        history['dialog'].extend(parse(obs['text'], splitSentences))

    history['episode_done'] = obs['episode_done']

    labels = obs.get('labels', obs.get('eval_labels', None))
    if labels is not None:
        if useStartEndIndices:
            history['labels'] = [dict.start_token + ' ' + l for l in labels]
        else:
            history['labels'] = labels

    return history['dialog']


def load_cands(path, lines_have_ids=False, cands_are_replies=False):
    """
    Load global fixed set of candidate labels that the teacher provides.

    Every example will include these as candidates. The true labels for a specific
    example are also added to this set, so that it's possible to get the right answer.
    """
    if path is None:
        return None
    cands = []
    cnt = 0
    with PathManager.open(path) as read:
        for line in read:
            line = line.strip().replace('\\n', '\n')
            if len(line) > 0:
                cnt = cnt + 1
                # If lines are numbered we strip them of numbers.
                if cnt == 1 and line[0:2] == '1 ':
                    lines_have_ids = True
                # If tabs then the label_candidates are all the replies.
                if '\t' in line and not cands_are_replies:
                    cands_are_replies = True
                    cands = []
                if lines_have_ids:
                    space_idx = line.find(' ')
                    line = line[space_idx + 1 :]
                    if cands_are_replies:
                        sp = line.split('\t')
                        if len(sp) > 1 and sp[1] != '':
                            cands.append(sp[1])
                    else:
                        cands.append(line)
                else:
                    cands.append(line)
    return cands


class Timer(object):
    """
    Computes elapsed time.
    """

    def __init__(self):
        """
        Initialize timer.
        """
        self.running = True
        self.total = 0
        self.start = time.time()

    def reset(self):
        """
        Reset timer to zero.
        """
        self.running = True
        self.total = 0
        self.start = time.time()
        return self

    def resume(self):
        """
        Resume timer.
        """
        if not self.running:
            self.running = True
            self.start = time.time()
        return self

    def stop(self):
        """
        Pause timer.
        """
        if self.running:
            self.running = False
            self.total += time.time() - self.start
        return self

    def time(self):
        """
        Get current timer time.
        """
        if self.running:
            return self.total + time.time() - self.start
        return self.total


class TimeLogger:
    """
    Class for logging time progress against a goal.
    """

    def __init__(self):
        """
        Set up timer.
        """
        self.timer = Timer()
        self.tot_time = 0

    def total_time(self):
        """
        Return time elapsed at last log call.
        """
        return self.tot_time

    def time(self):
        """
        Return current timer time.
        """
        return self.timer.time()

    def log(self, done, total, report=None):
        """
        Log report, time elapsed, and percentage progress towards goal.

        :param done: number of examples completed so far
        :param total: total number of elements to be completed. if total > 0,
                      calculates the time remaining and percentage complete.
        :param report: dict of pairs to log

        :returns: tuple log string, log dict
            log string contains time elapsed and string representation of
            the log dict
            log dict contains pairs of all items to log, which includes
            percentage complete and projected time left if total > 0
        """
        from parlai.core.metrics import Metric  # delay import to prevent circular dep

        if isinstance(done, Metric):
            done = done.value()
        self.tot_time += self.timer.time()
        self.timer.reset()
        if report:
            report['exs'] = done
        if total > 0 and done > 0:
            progress = done / total
            seconds_left = max(0, self.tot_time / progress - self.tot_time)
            eta = timedelta(seconds=int(seconds_left + 0.5))
        else:
            progress = 0
            eta = "unknown"
        elapsed = timedelta(seconds=int(self.tot_time))

        text = (
            f'{progress:.1%} complete ({done:,d} / {total:,d}), '
            f'{elapsed} elapsed, {eta} eta'
        )
        if report:
            report_s = nice_report(report)
            text = f'{text}\n{report_s}'
        return text, report


class AttrDict(dict):
    """
    Helper class to have a dict-like object with dot access.

    For example, instead of `d = {'key': 'value'}` use
    `d = AttrDict(key='value')`.
    To access keys, instead of doing `d['key']` use `d.key`.

    While this has some limitations on the possible keys (for example, do not
    set the key `items` or you will lose access to the `items()` method), this
    can make some code more clear.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize AttrDict using input dict.
        """
        super().__init__(*args, **kwargs)
        self.__dict__ = self


class SimpleCounter:
    """
    Simple counter object.
    """

    def __init__(self, value=0):
        self.val = value

    def increment(self, value=1):
        self.val += value

    def value(self):
        return self.val


def _report_sort_key(report_key: str) -> Tuple[str, str]:
    """
    Sorting name for reports.

    Sorts by main metric alphabetically, then by task.
    """
    # if metric is on its own, like "f1", we will return ('', 'f1')
    # if metric is from multitask, we denote it.
    # e.g. "convai2/f1" -> ('convai2', 'f1')
    # we handle multiple cases of / because sometimes teacher IDs have
    # filenames.
    fields = report_key.split("/")
    main_key = fields.pop(-1)
    sub_key = '/'.join(fields)
    return (sub_key or 'all', main_key)


def float_formatter(f: Union[float, int]) -> str:
    """
    Format a float as a pretty string.
    """
    if f != f:
        # instead of returning nan, return "" so it shows blank in table
        return ""
    if isinstance(f, int):
        # don't do any rounding of integers, leave them alone
        return str(f)
    if f >= 1000:
        # numbers > 1000 just round to the nearest integer
        s = f'{f:.0f}'
    else:
        # otherwise show 4 significant figures, regardless of decimal spot
        s = f'{f:.4g}'
    # replace leading 0's with blanks for easier reading
    # example:  -0.32 to -.32
    s = s.replace('-0.', '-.')
    if s.startswith('0.'):
        s = s[1:]
    # Add the trailing 0's to always show 4 digits
    # example: .32 to .3200
    if s[0] == '.' and len(s) < 5:
        s += '0' * (5 - len(s))
    return s


def _line_width():
    if os.environ.get('PARLAI_FORCE_WIDTH'):
        try:
            return int(os.environ['PARLAI_FORCE_WIDTH'])
        except ValueError:
            pass
    try:
        # if we're in an interactive ipython notebook, hardcode a longer width
        __IPYTHON__
        return 128
    except NameError:
        return shutil.get_terminal_size((88, 24)).columns


def nice_report(report) -> str:
    """
    Render an agent Report as a beautiful string.

    If pandas is installed,  we will use it to render as a table. Multitask
    metrics will be shown per row, e.g.

    .. code-block:
                 f1   ppl
       all     .410  27.0
       task1   .400  32.0
       task2   .420  22.0

    If pandas is not available, we will use a dict with like-metrics placed
    next to each other.
    """
    if not report:
        return ""

    from parlai.core.metrics import Metric

    try:
        import pandas as pd

        use_pandas = True
    except ImportError:
        use_pandas = False

    sorted_keys = sorted(report.keys(), key=_report_sort_key)
    output: OrderedDict[Union[str, Tuple[str, str]], float] = OrderedDict()
    for k in sorted_keys:
        v = report[k]
        if isinstance(v, Metric):
            v = v.value()
        if use_pandas:
            output[_report_sort_key(k)] = v
        else:
            output[k] = v

    if use_pandas:
        line_width = _line_width()

        df = pd.DataFrame([output])
        df.columns = pd.MultiIndex.from_tuples(df.columns)
        df = df.stack().transpose().droplevel(0, axis=1)
        result = "   " + df.to_string(
            na_rep="",
            line_width=line_width - 3,  # -3 for the extra spaces we add
            float_format=float_formatter,
            index=df.shape[0] > 1,
        ).replace("\n\n", "\n").replace("\n", "\n   ")
        result = re.sub(r"\s+$", "", result)
        return result
    else:
        return json.dumps(
            {
                k: round_sigfigs(v, 4) if isinstance(v, float) else v
                for k, v in output.items()
            }
        )


def round_sigfigs(x: Union[float, 'torch.Tensor'], sigfigs=4) -> float:
    """
    Round value to specified significant figures.

    :param x: input number
    :param sigfigs: number of significant figures to return

    :returns: float number rounded to specified sigfigs
    """
    x_: float
    if __TORCH_AVAILABLE and isinstance(x, torch.Tensor):
        x_ = x.item()
    else:
        x_ = x  # type: ignore

    try:
        if x_ == 0:
            return 0
        return round(x_, -(math.floor(math.log10(abs(x_)) - sigfigs + 1)))
    except (ValueError, OverflowError) as ex:
        if x_ in [float('inf'), float('-inf')] or x_ != x_:  # inf or nan
            return x_
        else:
            raise ex


def clip_text(text, max_len):
    """
    Clip text to max length, adding ellipses.
    """
    if len(text) > max_len:
        begin_text = ' '.join(text[: math.floor(0.8 * max_len)].split(' ')[:-1])
        end_text = ' '.join(
            text[(len(text) - math.floor(0.2 * max_len)) :].split(' ')[1:]
        )
        if len(end_text) > 0:
            text = begin_text + ' ...\n' + end_text
        else:
            text = begin_text + ' ...'
    return text


def _ellipse(lst: List[str], max_display: int = 5, sep: str = '|') -> str:
    """
    Like join, but possibly inserts an ellipsis.

    :param lst: The list to join on
    :param int max_display: the number of items to display for ellipsing.
        If -1, shows all items
    :param string sep: the delimiter to join on
    """
    # copy the list (or force it to a list if it's a set)
    choices = list(lst)
    # insert the ellipsis if necessary
    if max_display > 0 and len(choices) > max_display:
        ellipsis = '... ({} of {} shown)'.format(max_display, len(choices))
        choices = choices[:max_display] + [ellipsis]
    return sep.join(str(c) for c in choices)


def display_messages(
    msgs: List[Dict[str, Any]],
    prettify: bool = False,
    ignore_agent_reply: bool = False,
    add_fields: str = '',
    max_len: int = 1000,
    verbose: bool = False,
) -> Optional[str]:
    """
    Return a string describing the set of messages provided.

    If prettify is true, candidates are displayed using prettytable. add_fields provides
    a list of fields in the msgs which should be displayed if verbose is off.
    """

    def _token_losses_line(
        msg: Dict[str, Any], fields_to_show: List[str], space: str
    ) -> Optional[str]:
        """
        Displays the loss associated with each token. Can be used for debugging
        generative models.

        See TorchGeneratorAgent._construct_token_losses for an example implementation.
        """
        key = 'token_losses'
        token_losses = msg.get(key, None)
        if key not in fields_to_show or not token_losses:
            return None
        # Reduce losses to 4 significant figures
        formatted_tl = ' | '.join(
            [f"{tl[0]} {float('{:.4g}'.format(tl[1]))}" for tl in token_losses]
        )
        return _pretty_lines(space, key, formatted_tl, 'text2')

    def _pretty_lines(indent_space, field, value, style):
        line = '{}{} {}'.format(
            indent_space, colorize('[' + field + ']:', 'field'), colorize(value, style)
        )
        return line

    lines = []
    episode_done = False
    extra_add_fields_ = add_fields.split(',')
    for index, msg in enumerate(msgs):
        if msg is None or (index == 1 and ignore_agent_reply):
            # We only display the first agent (typically the teacher) if we
            # are ignoring the agent reply.
            continue

        if msg.get('episode_done'):
            episode_done = True
        # Possibly indent the text (for the second speaker, if two).
        space = ''
        if len(msgs) == 2 and index == 1:
            space = '   '

        agent_id = msg.get('id', '[no id field]')
        if verbose:
            line = _pretty_lines(
                indent_space=space, field='id', value=agent_id, style='id'
            )
            lines.append(line)

        # Only display rewards !=0 as they are confusing in non-RL tasks.
        if msg.get('reward', 0) != 0:
            lines.append(space + '[reward: {r}]'.format(r=msg['reward']))

        fields_to_show = []
        if verbose:
            fields_to_show = [field for field in msg]
        else:
            fields_to_show = [
                field
                for field in msg
                if field in list(MUST_SHOW_MESSAGE_FIELDS) + extra_add_fields_
            ]
        fields_to_show.sort()

        # Display fields without special format
        for field in fields_to_show:
            if field not in SPECIAL_FORMATED_DISPLAY_MESSAGE_FIELDS:
                if type(msg[field]) is list:
                    value = _ellipse(msg[field], sep='\n  ')
                else:
                    value = clip_text(str(msg.get(field)), max_len)
                line = _pretty_lines(
                    indent_space=space, field=field, value=value, style='text2'
                )
                lines.append(line)

        # Display fields WITH special format requirements
        # Display Image
        if type(msg.get('image')) in [str, torch.Tensor]:
            lines.append(f'[ image ]: {msg["image"]}')
        # Display Text
        if msg.get('text', ''):
            value = clip_text(msg['text'], max_len)
            style = 'bold_text' if index == 0 else 'labels'
            field = 'text' if verbose else agent_id
            line = _pretty_lines(
                indent_space=space, field=field, value=value, style=style
            )
            lines.append(line)
        # Display Label Fields
        for field in {'labels', 'eval_labels', 'label_candidates', 'text_candidates'}:
            if msg.get(field) and field in fields_to_show:
                line = _pretty_lines(
                    indent_space=space,
                    field=field,
                    value=_ellipse(msg[field]),
                    style=field,
                )
                lines.append(line)
        if msg.get('metrics') and verbose:
            lines.append(
                _pretty_lines(
                    indent_space=space,
                    field='metrics',
                    value="\n" + nice_report(msg['metrics']),
                    style='text',
                )
            )

        # Handling this separately since we need to clean up the raw output before displaying.
        token_loss_line = _token_losses_line(msg, fields_to_show, space)
        if token_loss_line:
            lines.append(token_loss_line)

    if episode_done:
        lines.append(
            colorize('- - - - - - - END OF EPISODE - - - - - - - - - -', 'highlight')
        )

    return '\n'.join(lines)


def str_to_msg(txt, ignore_fields=''):
    """
    Convert formatted string to ParlAI message dict.

    :param txt:
        formatted string to convert. String format is tab-separated fields,
        with colon separating field name and contents.
    :param ignore_fields:
        (default '') comma-separated field names to not
        include in the msg dict even if they're in the string.
    """

    def tostr(txt):
        txt = str(txt)
        txt = txt.replace('\\t', '\t')
        txt = txt.replace('\\n', '\n')
        txt = txt.replace('__PIPE__', '|')
        return txt

    def tolist(txt):
        vals = txt.split('|')
        for v in vals:
            v = tostr(v)
        return vals

    def convert(key, value):
        if key == 'text' or key == 'id':
            return tostr(value)
        elif (
            key == 'label_candidates'
            or key == 'labels'
            or key == 'eval_labels'
            or key == 'text_candidates'
        ):
            return tolist(value)
        elif key == 'episode_done':
            return bool(value)
        else:
            return tostr(value)

    if txt == '' or txt is None:
        return None

    msg = {}
    for t in txt.split('\t'):
        ind = t.find(':')
        key = t[:ind]
        value = t[ind + 1 :]
        if key not in ignore_fields.split(','):
            msg[key] = convert(key, value)
    msg['episode_done'] = msg.get('episode_done', False)
    return Message(msg)


def msg_to_str(msg, ignore_fields=''):
    """
    Convert ParlAI message dict to string.

    :param msg:
        dict to convert into a string.
    :param ignore_fields:
        (default '') comma-separated field names to not include in the string
        even if they're in the msg dict.
    """

    def filter(txt):
        txt = str(txt)
        txt = txt.replace('\t', '\\t')
        txt = txt.replace('\n', '\\n')
        txt = txt.replace('|', '__PIPE__')
        return txt

    def add_field(name, data):
        if name == 'reward' and data == 0:
            return ''
        if name == 'episode_done' and data is False:
            return ''
        txt = ''
        if type(data) == tuple or type(data) == set or type(data) == list:
            # list entries
            for c in data:
                txt += filter(c) + "|"
            txt = txt[:-1]
        else:
            # single fields
            txt = filter(data)
        return name + ":" + txt + '\t'

    default_fields = [
        'id',
        'text',
        'labels',
        'label_candidates',
        'episode_done',
        'reward',
    ]
    txt = ""
    ignore_fields = ignore_fields.split(',')
    for f in default_fields:
        if f in msg and f not in ignore_fields:
            txt += add_field(f, msg[f])
    for f in msg.keys():
        if f not in default_fields and f not in ignore_fields:
            txt += add_field(f, msg[f])
    return txt.rstrip('\t')


# DEPRECATION DAY: DELETE
def set_namedtuple_defaults(namedtuple, default=None):
    """
    Set *all* of the fields for a given nametuple to a singular value.

    Additionally removes the default docstring for each field.
    Modifies the tuple in place, but returns it anyway.

    More info:
    https://stackoverflow.com/a/18348004

    :param namedtuple: A constructed collections.namedtuple
    :param default: The default value to set.

    :returns: the modified namedtuple
    """
    namedtuple.__new__.__defaults__ = (default,) * len(namedtuple._fields)
    for f in namedtuple._fields:
        del getattr(namedtuple, f).__doc__
    return namedtuple


_seen_logs: Set[str] = set()


def warn_once(msg: str) -> None:
    """
    Log a warning, but only once.

    :param str msg: Message to display
    """
    global _seen_logs
    if msg not in _seen_logs:
        _seen_logs.add(msg)
        logging.warn(msg)


def error_once(msg: str) -> None:
    """
    Log an error, but only once.

    :param str msg: Message to display
    """
    global _seen_logs
    if msg not in _seen_logs:
        _seen_logs.add(msg)
        logging.error(msg)


def recursive_getattr(obj, attr, *args):
    """
    Recursive call to getattr for nested attributes.
    """

    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return functools.reduce(_getattr, [obj] + attr.split('.'))



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Basic tokenizer that splits text into alpha-numeric tokens and non-whitespace tokens.
"""

import regex
from .tokenizer import Tokens, Tokenizer
from parlai.utils.logging import logger


class SimpleTokenizer(Tokenizer):
    ALPHA_NUM = r'[\p{L}\p{N}\p{M}]+'
    NON_WS = r'[^\p{Z}\p{C}]'

    def __init__(self, **kwargs):
        """
        Args:
            annotators: None or empty set (only tokenizes).
        """
        self._regexp = regex.compile(
            '(%s)|(%s)' % (self.ALPHA_NUM, self.NON_WS),
            flags=regex.IGNORECASE + regex.UNICODE + regex.MULTILINE,
        )
        if len(kwargs.get('annotators', {})) > 0:
            logger.warning(
                '%s only tokenizes! Skipping annotators: %s'
                % (type(self).__name__, kwargs.get('annotators'))
            )
        self.annotators = set()

    def tokenize(self, text):
        data = []
        matches = [m for m in self._regexp.finditer(text)]
        for i in range(len(matches)):
            # Get text
            token = matches[i].group()

            # Get whitespace
            span = matches[i].span()
            start_ws = span[0]
            if i + 1 < len(matches):
                end_ws = matches[i + 1].span()[0]
            else:
                end_ws = span[1]

            # Format data
            data.append((token, text[start_ws:end_ws], span))
        return Tokens(data, self.annotators)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
import copy
import json
import os
import re

import numpy as np

from parlai.core.teachers import ParlAIDialogTeacher, MultiTaskTeacher
from projects.self_feeding.utils import add_person_tokens
from parlai.utils.io import PathManager
from .build import build


def _path(opt, filename, add_suffix=False):
    build(opt)
    if add_suffix:
        subtask = opt['subtask']
        if subtask == 'dialog':
            suffix = 'hh'
        elif subtask == 'satisfaction':
            suffix = 'st'
        elif subtask == 'feedback':
            suffix = 'fb'
        else:
            raise ValueError(f"Unrecognized subtask: {subtask}")
        filename += f"_{suffix}.txt"
    dp = os.path.join(opt['datapath'], 'self_feeding')
    return os.path.join(dp, filename)


class SelfFeedingTeacher(ParlAIDialogTeacher):
    """
    Teacher for the SelfFeedingAgent.

    opt['datatype'] determines whether we use the designated filepath ('train') or one
    of the eval files ('valid', 'test'), which are identical regardless of     what
    training set is being used.
    """

    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        if 'subtask' not in opt:
            print(
                'Warning: SelfFeedingteacher should be assigned subtask. '
                'Defaulting to dialog'
            )
            opt['subtask'] = 'dialog'

        # Use 'in' to also capture 'train:ordered:stream'
        if 'train' in opt['datatype']:
            # Use the filename explicitly given with the flag if available
            # Otherwise, use train_xx.txt where xx is inferred from the subtask
            train_file_flag = f"{opt['subtask'][:3]}_train"
            if opt.get(train_file_flag, None):
                path = _path(opt, opt[train_file_flag], add_suffix=False)
            else:
                path = _path(opt, "train", add_suffix=True)
        else:
            # Use the filename explicitly given with the flag if available
            # Otherwise, use the datatype (valid_xx.txt or test_xx.txt) where xx is
            # inferred from the subtask.
            eval_file_flag = f"{opt['subtask'][:3]}_{opt['datatype']}"
            if opt.get(eval_file_flag, None):
                path = _path(opt, opt[eval_file_flag], add_suffix=False)
            else:
                path = _path(opt, opt['datatype'].split(':')[0], add_suffix=True)

        if not PathManager.exists(path):
            raise ValueError("Unrecognized filepath: {}".format(path))

        opt['parlaidialogteacher_datafile'] = path
        opt['datafile'] = path
        super().__init__(opt, shared)

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        project = parser.add_argument_group('Self-Feeding Tasks')
        project.add_argument(
            '-st',
            '--subtasks',
            type=str,
            help='comma-separated list of tasks used by MTL teacher',
        )
        project.add_argument(
            '-dia-train',
            '--dia-train',
            type=str,
            help='the filename to train on for the dialog task',
        )
        project.add_argument(
            '-fee-train',
            '--fee-train',
            type=str,
            help='the filename to train on for the feedback task',
        )
        project.add_argument(
            '-sat-train',
            '--sat-train',
            type=str,
            help='the filename to train on for the satisfaction task',
        )
        project.add_argument(
            '-dia-valid',
            '--dia-valid',
            type=str,
            help='the filename to eval on for the dialog task',
        )
        project.add_argument(
            '-fee-valid',
            '--fee-valid',
            type=str,
            help='the filename to eval on for the feedback task',
        )
        project.add_argument(
            '-sat-valid',
            '--sat-valid',
            type=str,
            help='the filename to eval on for the satisfaction task',
        )
        project.add_argument(
            '-dia-test',
            '--dia-test',
            type=str,
            help='the filename to eval on for the dialog task',
        )
        project.add_argument(
            '-fee-test',
            '--fee-test',
            type=str,
            help='the filename to eval on for the feedback task',
        )
        project.add_argument(
            '-sat-test',
            '--sat-test',
            type=str,
            help='the filename to eval on for the satisfaction task',
        )
        project.add_argument(
            '-trial',
            '--trial',
            type=int,
            default=0,
            help='the index of a repeated trial (not used in code)',
        )
        project.add_argument(
            '-mt',
            '--max-train',
            type=int,
            default=0,
            help='if non-zero, only the first max-train examples will '
            'be used if it is read by an instance of '
            'ParlaiDialogTeacher',
        )
        parser.set_defaults(history_size=2)
        return parser

    def _setup_data(self, path):  # Make private method for ParlAIDialogTeacher
        """
        Reads data in the fbdialog format.

        Returns ``((x,y,r,c), new_episode?)`` tuples.
        """
        print("[ Loading Self-Feeding text data:" + path + "]")
        self.episodes = []
        self.num_exs = 0
        self.max_train = self.opt.get('max_train', 0)
        with PathManager.open(path, 'r') as f:
            for line in f.readlines():
                if self.max_train and self.num_exs >= self.max_train:
                    break
                parley = json.loads(line)

                # NOTE: History is trimmed here, not by TorchAgent (except in
                # interactive mode)
                if self.opt['history_size'] == 0:
                    parley['context'] = '__null__'
                elif self.opt['history_size'] > 0:
                    utterances = re.split(r'__p\d__', parley['context'])[1:]
                    trimmed = utterances[-(self.opt['history_size']) :]
                    parley['context'] = add_person_tokens(trimmed, last_speaker=1)

                # WARNING: STRIPPING AWAY MEMORIES
                parley['memories'] = []

                episode = {
                    'text': parley['context'],
                    'labels': [parley['response']],
                    'label_candidates': parley.get('candidates', []),
                    'reward': parley.get('reward', 0),
                    'episode_done': True,
                }

                # Convert integer labels (e.g., polarization dataset) to strings
                episode['labels'] = [str(l) for l in episode['labels']]

                self.num_exs += 1
                self.episodes.append([episode])


class SelfFeedingMTLTeacher(MultiTaskTeacher):
    """
    Creates a teacher that is actually a set of teachers each based on a task string--
    each of these teachers will get called in turn, either randomly or in order. They
    are all in the same world (they are the same agent switching tasks).

    More specifically, this child class of MultiTaskTeacher supports multitask learning
    with batches (ensuring that all batches only have data from a single task at a time)
    """

    def __init__(self, opt, shared=None):
        if opt['task'] == 'self_feeding:SelfFeedingMTLTeacher':
            # This does not happen except in tests such as test_new_tasks.py
            opt = copy.deepcopy(opt)
            opt['task'] = 'self_feeding:SelfFeedingTeacher'
        super().__init__(opt, shared)

        # TODO: allow user to specify other strategies for creating sampling_prob
        # uniform, batch_proportion, etc.
        num_batches = np.array([t.num_examples() / t.bsz for t in self.tasks])
        self.sampling_prob = num_batches / np.sum(num_batches)

        self.task_idx_assignment = -1  # This will be updated by BatchWorld
        self.new_task = True
        self.random = opt.get('datatype') == 'train'

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        SelfFeedingTeacher.add_cmdline_args(parser, partial_opt=partial_opt)
        return parser

    def observe(self, observation):
        return self.tasks[self.task_idx].observe(observation)

    def batch_act(self, batch_observation):
        # We use only one task for each batch
        task_idx = self.get_task_index()
        actions = []
        for _ in range(self.tasks[task_idx].bsz):
            action = self.tasks[task_idx].act()
            # Pass the name of the task currently being worked on
            action['subtask'] = self.tasks[task_idx].opt['subtask']
            actions.append(action)
        return actions

    def act(self):
        self.task_idx = self.get_task_index()
        if self.task_idx < 0:
            return {'episode_done': True}
        action = self.tasks[self.task_idx].act()
        # Pass the name of the task currently being worked on
        action['subtask'] = self.tasks[self.task_idx].opt['subtask']
        return action

    def get_task_index(self):
        if self.opt['datatype'] == 'train':
            return np.random.choice(range(len(self.tasks)), p=self.sampling_prob)
        else:
            for i, subtask in enumerate(self.tasks):
                if not subtask.epoch_done():
                    return i
        # If this is reached, all tasks are done, so return sentinel
        return -1


class DialogTeacher(SelfFeedingTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['subtask'] = 'dialog'
        super().__init__(opt, shared)

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt=partial_opt)
        return parser


class FeedbackTeacher(SelfFeedingTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['subtask'] = 'feedback'
        super().__init__(opt, shared)

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt=partial_opt)
        return parser


class SatisfactionTeacher(SelfFeedingTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['subtask'] = 'satisfaction'
        super().__init__(opt, shared)

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt=partial_opt)
        return parser


class DiafeeTeacher(SelfFeedingMTLTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['subtasks'] = ['dialog', 'feedback']
        # Expand abbreviated task name ('both') into full task names
        train_files = [opt['dia_train'], opt['fee_train']]
        assert len(opt['subtasks']) == len(train_files)
        tasks = [
            f'self_feeding:{subtask}:{train_file}'
            for subtask, train_file in zip(opt['subtasks'], train_files)
        ]
        opt['task'] = ','.join(tasks)
        super().__init__(opt, shared)

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        SelfFeedingTeacher.add_cmdline_args(parser, partial_opt=partial_opt)
        return parser


class DiasatTeacher(SelfFeedingMTLTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['subtasks'] = ['dialog', 'satisfaction']
        # Expand abbreviated task name ('both') into full task names
        train_files = [opt['dia_train'], opt['sat_train']]
        assert len(opt['subtasks']) == len(train_files)
        tasks = [
            f'self_feeding:{subtask}:{train_file}'
            for subtask, train_file in zip(opt['subtasks'], train_files)
        ]
        opt['task'] = ','.join(tasks)
        super().__init__(opt, shared)

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        SelfFeedingTeacher.add_cmdline_args(parser, partial_opt=partial_opt)
        return parser


class AllTeacher(SelfFeedingMTLTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['subtasks'] = ['dialog', 'feedback', 'satisfaction']
        # Expand abbreviated task name ('all') into full task names
        train_files = [opt['dia_train'], opt['fee_train'], opt['sat_train']]
        assert len(opt['subtasks']) == len(train_files)
        tasks = [
            f'self_feeding:{subtask}:{train_file}'
            for subtask, train_file in zip(opt['subtasks'], train_files)
        ]
        opt['task'] = ','.join(tasks)
        super().__init__(opt, shared)

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        SelfFeedingTeacher.add_cmdline_args(parser, partial_opt=partial_opt)
        return parser



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

from parlai.core.build_data import DownloadableFile
import parlai.core.build_data as build_data
import os

RESOURCES = [
    DownloadableFile(
        'https://storage.googleapis.com/dialog-data-corpus/CCPE-M-2019/data.json',
        'ccpe.json',
        '4ff051ea7ea60cf0f480c911c7e2cfed56434e2e2c9ea8965ac5e26365773f0a',
        zipped=False,
    )
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'CCPE')
    version = '1.0'

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Pretrained models used by unit tests.
"""

import os
from parlai.core.build_data import download_models, built, get_model_dir


def download(datapath):
    opt = {'datapath': datapath}
    model_name = 'unittest'
    mdir = os.path.join(get_model_dir(datapath), model_name)
    version = 'v7.1'
    model_filenames = [
        'seq2seq.tar.gz',
        'transformer_ranker.tar.gz',
        'transformer_generator2.tar.gz',
        'memnn.tar.gz',
        'apex_v1.tar.gz',
        'test_bytelevel_bpe_v2.tar.gz',
        'beam_blocking1.tar.gz',
        'context_blocking1.tar.gz',
        'hred_model_v1.tar.gz',
    ]
    if not built(mdir, version):
        download_models(opt, model_filenames, model_name, version=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.utils.testing import AutoTeacherTest


class TestDefaultTeacher(AutoTeacherTest):
    task = "blended_skill_talk"


class TestBlendedSkillTalkTeacher(AutoTeacherTest):
    task = "blended_skill_talk:blended_skill_talk"


class TestConvAI2PersonaTopicifierTeacher(AutoTeacherTest):
    task = "blended_skill_talk:convAI2_persona_topicifier"


class TestWoWPersonaTopicifierTeacher(AutoTeacherTest):
    task = "blended_skill_talk:wo_w_persona_topicifier"


class TestEDPersonaTopicifierTeacher(AutoTeacherTest):
    task = "blended_skill_talk:e_d_persona_topicifier"


class TestAllTeacher(AutoTeacherTest):
    task = "blended_skill_talk:all"



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import FbDeprecatedDialogTeacher
from .build import build

import copy
import os


def _path(opt):
    build(opt)
    suffix = ''
    dt = opt['datatype'].split(':')[0]
    if dt == 'train':
        suffix = 'train'
    elif dt == 'test':
        suffix = 'test'
    elif dt == 'valid':
        suffix = 'dev'
    return os.path.join(
        opt['datapath'],
        'WikiMovies',
        'movieqa',
        'questions',
        'wiki_entities',
        'wiki-entities_qa_{suffix}.txt'.format(suffix=suffix),
    )


# The knowledge base of facts that can be used to answer questions.
class KBTeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        build(opt)
        task = opt.get('task')
        if not task:
            task = 'wikimovies:KB:kb'
        kb = task.split(':')
        if len(kb) == 3:
            kb = kb[2]
        elif len(kb) == 2:
            # default to 'kb' if 'kb', 'wiki', or 'ie' not specified
            kb = 'kb'
        kbs = {}
        kbs['kb'] = os.path.join('wiki_entities', 'wiki_entities_kb.txt')
        kbs['wiki'] = 'wiki.txt'
        kbs['ie'] = 'wiki_ie.txt'
        opt['datafile'] = os.path.join(
            opt['datapath'], 'WikiMovies', 'movieqa', 'knowledge_source', kbs[kb]
        )
        super().__init__(opt, shared)


class DefaultTeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        build(opt)
        opt = copy.deepcopy(opt)
        opt['datafile'] = _path(opt)
        opt['cands_datafile'] = os.path.join(
            opt['datapath'], 'WikiMovies', 'movieqa', 'knowledge_source', 'entities.txt'
        )
        super().__init__(opt, shared)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.


import os
from parlai.core.build_data import DownloadableFile
import parlai.core.build_data as build_data
import parlai.utils.logging as logging

RESOURCES = [
    DownloadableFile(
        'http://parl.ai/downloads/convai2/convai2_fix_723.tgz',
        'convai2_fix_723.tgz',
        'd0ae89defe2fd0b0a4221eaa642a457d7d40cef475f54798119c7f3b8dd9361d',
    )
]


def build(opt):
    version = 'v5.0'
    dpath = os.path.join(opt['datapath'], 'ConvAI2')

    if not build_data.built(dpath, version):
        logging.info('building data: ' + dpath)
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import FixedDialogTeacher
from parlai.utils.io import PathManager
from .build import build
import os
import json
import re
import csv


def _path(opt):
    # build the data if it does not exist
    build(opt)

    # set up path to data (specific to each dataset)
    data_path = os.path.join(opt['datapath'], 'redial')
    return data_path


# Turns title from format "Title (Year)" to "Title" or leaves as is if no (Year)
def remove_year_from_title(title):
    matches = re.finditer(r"\s\(", title)
    indices = [m.start(0) for m in matches]
    if indices:
        title_end = indices[-1]
        return title[:title_end]
    else:
        return title


def replace_movie_ids(id_string, id_map):
    pattern = r'@\d+'
    return re.sub(pattern, lambda s: id_map[s.group()], id_string)


class ReDialTeacher(FixedDialogTeacher):
    """
    ReDial Teacher.
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        data_path = _path(opt)
        self.title_id_map = {}
        self.get_title_dict(data_path)
        if shared is not None:
            self.episodes = shared['episodes']
        else:
            self.episodes = []
            self._setup_data(data_path)
        self.id = 'redial'

        self.reset()

    def get_title_dict(self, path):
        csv_path = os.path.join(path, 'movies_with_mentions.csv')
        with PathManager.open(csv_path, mode='r') as f:
            reader = csv.reader(f)
            for row in reader:
                self.title_id_map['@' + row[0]] = remove_year_from_title(row[1])

    def _setup_data(self, data_path):
        train_path = os.path.join(data_path, 'train_data.jsonl')
        test_path = os.path.join(data_path, 'test_data.jsonl')
        # The test data has 1341 episodes. Making valid this size gives
        # about 80/10/10 train/test/valid split
        test_set_episodes = 1341
        if self.datatype.startswith('test'):
            unmerged_episodes = self.get_data_from_file(test_path)
        elif self.datatype.startswith('valid'):
            unmerged_episodes = self.get_data_from_file(train_path)
            unmerged_episodes = unmerged_episodes[:test_set_episodes]
        else:
            unmerged_episodes = self.get_data_from_file(train_path)
            unmerged_episodes = unmerged_episodes[test_set_episodes:]

        # some speakers speak multiple times in a row.
        for unmerged_episode in unmerged_episodes:
            episode = []
            prev_speaker = None
            for message in unmerged_episode['messages']:
                curr_speaker = message['senderWorkerId']
                text = replace_movie_ids(message['text'], self.title_id_map)
                if curr_speaker == prev_speaker:
                    episode[-1] = episode[-1] + " " + text
                else:
                    episode.append(text)
                    prev_speaker = curr_speaker
            self.episodes.append(episode)

    def get_data_from_file(self, filepath):
        data = []
        with PathManager.open(filepath) as f:
            for line in f:
                data.append(json.loads(line))
        return data

    def share(self):
        shared = super().share()
        shared['episodes'] = self.episodes
        return shared

    def num_examples(self):
        examples = 0
        for data in self.episodes:
            examples += len(data) // 2
        return examples

    def num_episodes(self):
        return len(self.episodes)

    def get(self, episode_idx, entry_idx=0):
        text_idx = entry_idx * 2
        entry = self.episodes[episode_idx][text_idx]
        final_speaker_idx = len(self.episodes[episode_idx]) - 2
        # sometimes the first speaker is at the end with no reply
        if len(self.episodes[episode_idx]) % 2 == 1:
            final_speaker_idx -= 1
        labels = [self.episodes[episode_idx][text_idx + 1]]
        episode_done = text_idx >= final_speaker_idx
        action = {
            'id': self.id,
            'text': entry,
            'episode_done': episode_done,
            'labels': labels,
        }
        return action


class DefaultTeacher(ReDialTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Simple agent which repeats back the last thing said to it.

Useful as a baseline for metrics like F1.
"""

from parlai.core.agents import Agent
from parlai.core.message import Message


class RepeatQueryAgent(Agent):
    def __init__(self, opt, shared=None):
        super().__init__(opt)
        self.id = 'RepeatQueryAgent'

    def act(self):
        obs = self.observation
        if obs is None:
            return Message({'text': 'Nothing to repeat yet.', 'episode_done': False})
        reply = {}
        reply['id'] = self.getID()
        query = obs.get('text', "I don't know")
        # Take last line if there are multiple lines.
        reply['text'] = query.split('\n')[-1]
        if reply['text'] == '':
            reply['text'] = 'Nothing to repeat yet.'
        reply['episode_done'] = False
        return Message(reply)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile

DECODE = 'decode'
DECODE_PREFIX = 'decode_'
DECODE_VERSION = 'v0.1'


RESOURCES = [
    DownloadableFile(
        'https://sharenlpfile-01.s3.amazonaws.com/data/decode_v0.1.zip',
        'decode_v0.1.zip',
        # '0badc03c41813ae9748f259370ce655e576d736fea2d084dd6a786ac59f2f2a1',
        '084aab98652a04ce4a78c1a63d91575f5ab416a0c474b962ca1f4508a56b7484',
        True,
        False,
    )
]


def build(opt):
    dpath = os.path.join(opt['datapath'], DECODE)

    version = DECODE_VERSION

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

import json
import os

from parlai.core import build_data
from parlai.utils.io import PathManager


RESOURCES = [
    build_data.DownloadableFile(
        'http://parl.ai/downloads/blended_skill_talk/blended_skill_talk.tar.gz',
        'blended_skill_talk.tar.gz',
        '5fbed0068ee89e2d43b93c3ecb341e784617033efa5e8e911a219d4eda6134a6',
    ),
    build_data.DownloadableFile(
        'http://parl.ai/downloads/blended_skill_talk/personas_list.txt',
        'persona_list.txt',
        '59a51adedc78e806a380f16477de3740cefe3494d20f8a2a733841bedaaa3ee5',
        zipped=False,
    ),
    build_data.DownloadableFile(
        'http://parl.ai/downloads/blended_skill_talk/topic_to_persona_list.txt',
        'topic_to_persona_list.txt',
        '47cdb6cbee0516ca7400be35fa07761339b86c6c026425bf5dba00e5534e8182',
        zipped=False,
    ),
    build_data.DownloadableFile(
        'http://parl.ai/downloads/blended_skill_talk/ed_persona_topicifier__train__both_sides.json',
        'ed_persona_topicifier__train__both_sides.json',
        'ff2ea7c5fcb0449890d57a629cc3e8794ab95ac6db1057bf58d540c2b576e4cc',
        zipped=False,
    ),
    build_data.DownloadableFile(
        'http://parl.ai/downloads/blended_skill_talk/ed_persona_topicifier__train__experiencer_only.json',
        'ed_persona_topicifier__train__experiencer_only.json',
        '751f0ba2f421a11eee2bfc896d60ab70d669093c3a5f6cb30e8d202133a90ec7',
        zipped=False,
    ),
    build_data.DownloadableFile(
        'http://parl.ai/downloads/blended_skill_talk/ed_persona_topicifier__valid__experiencer_only.json',
        'ed_persona_topicifier__valid__experiencer_only.json',
        '15d5412f5990a8a9c892305009d8597a737322aafe878b03ec71143703b25ba0',
        zipped=False,
    ),
    build_data.DownloadableFile(
        'http://parl.ai/downloads/blended_skill_talk/ed_persona_topicifier__test__experiencer_only.json',
        'ed_persona_topicifier__test__experiencer_only.json',
        '2604e977787be0b5edc54561f7ce8a54c40758d235a3fee262fe20fe36b8cd15',
        zipped=False,
    ),
    build_data.DownloadableFile(
        'http://parl.ai/downloads/blended_skill_talk/safe_personas_2.txt',
        'safe_personas.txt',
        '2ee292aa0006ea002e9b23d4f7326fe9e17514ce5793d31fd8d679035d4366a7',
        zipped=False,
    ),
    build_data.DownloadableFile(
        'http://parl.ai/downloads/blended_skill_talk/human_annotations.json',
        'human_annotations.json',
        'fe76a989d4fdcfa4a5beb214054069ed0abf45f3530eaeb8df38e75f7bb14795',
        zipped=False,
    ),
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'blended_skill_talk')
    version = 'v1.5'

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # Format it for use with ParlAIDialogTeacher
        _create_parlai_format(dpath)

        # Mark the data as built
        build_data.mark_done(dpath, version_string=version)


def _create_parlai_format(dpath: str):
    """
    Copy data into the format read by ParlAIDialogTeacher.

    'text' will be from the free Turker, who speaks first, and 'label' will be from the
    guided Turker.
    """

    datatypes = ['train', 'valid', 'test']
    for datatype in datatypes:

        load_path = os.path.join(dpath, f'{datatype}.json')
        save_path = os.path.join(dpath, f'{datatype}.txt')

        print(f'Loading {load_path}.')
        with PathManager.open(load_path, 'r', encoding='utf8') as f_read:
            data = json.load(f_read)

        print(f'Saving to {save_path}')
        with PathManager.open(save_path, 'w', encoding='utf8') as f_write:
            for episode in data:
                assert (
                    len(episode['dialog'])
                    == len(episode['suggestions'])
                    == len(episode['chosen_suggestions'])
                )
                num_entries = len(episode['dialog']) // 2
                for entry_idx in range(num_entries):
                    line = _get_line(
                        episode=episode, num_entries=num_entries, entry_idx=entry_idx
                    )
                    f_write.write(f'{line} \n')


def _get_line(episode: dict, num_entries: int, entry_idx: int) -> str:
    """
    Return the line to print in the reformatted file.
    """
    episode_done = entry_idx == num_entries - 1

    # Compile original context
    if entry_idx == 0:
        # Add those pieces of context that appear in the datasets that this one was
        # based on. Specifically:
        # - Your persona, but not your partner's persona (from ConvAI2)
        # - Topic (from Wizard of Wikipedia)
        # - **Not** the situation (from EmpatheticDialogues)
        persona_pieces = [
            f"your persona: {episode['personas'][1][0]}",
            f"your persona: {episode['personas'][1][1]}",
        ]
        if episode['context_dataset'] == 'wizard_of_wikipedia':
            additional_context_pieces = [episode['additional_context']]
        else:
            additional_context_pieces = []
        previous_utterance_pieces = [
            episode['free_turker_utterance'],
            episode['guided_turker_utterance'],
        ]
        original_context = (
            '\n'.join(
                persona_pieces + additional_context_pieces + previous_utterance_pieces
            )
            + '\n'
        )
    else:
        original_context = ''

    # Gather messages and suggestions
    free_message = episode['dialog'][2 * entry_idx][1]
    guided_message = episode['dialog'][2 * entry_idx + 1][1]
    single_task_suggestions = {
        task: episode['suggestions'][2 * entry_idx + 1][task]
        for task in ['convai2', 'empathetic_dialogues', 'wizard_of_wikipedia']
    }
    guided_chosen_suggestion = episode['chosen_suggestions'][2 * entry_idx + 1]

    # Compile into text string
    parts = {
        'text': original_context + free_message,
        'labels': guided_message,
        'context_dataset': episode['context_dataset'],
        'free_message': free_message,
        **single_task_suggestions,
        'guided_chosen_suggestion': guided_chosen_suggestion,
    }
    assert all([isinstance(part, str) for part in parts.values()])
    line = '\t'.join([f'{key}:{_escape(value)}' for key, value in parts.items()])

    # Add episode_done
    if episode_done:
        line += '\tepisode_done:True'

    # Add label_candidates
    if 'label_candidates' in episode:
        label_candidates = episode['label_candidates'][entry_idx]
        # Note that episode['dialog'] is indexed by utterance (from either Turker) and
        # episode['label_candidates'] is indexed by guided Turker response
        assert all([isinstance(cand, str) for cand in label_candidates])
        escaped_label_candidates = [_escape(cand) for cand in label_candidates]
        line += '\tlabel_candidates:' + '|'.join(escaped_label_candidates)
    return line


def _escape(value: str) -> str:
    return value.replace('\t', '\\t').replace('\n', '\\n').replace('|', '__PIPE__')



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Miscellaneous utils for chat services.
"""
import importlib
import math
import re
from enum import Enum

THREAD_MEDIUM_SLEEP = 0.3


class TaskState:
    """
    Wrapper for an agent running on a Worker.
    """

    def __init__(
        self, task_name, world_name, agents, is_overworld=False, world_type=None
    ):
        self.task_name = task_name
        self.world_name = world_name
        self.agents = agents
        self.is_overworld = is_overworld  # (bool): overworld or task world
        self.world_type = world_type  # name of the task world returned by the overworld

        self.future = None
        self.world = None  # world object


def get_world_module(world_path):
    """
    Import the module specified by the world_path.
    """
    run_module = None
    try:
        run_module = importlib.import_module(world_path)
    except Exception as e:
        print("Could not import world file {}".format(world_path))
        raise e
    return run_module


def get_world_fn_attr(world_module, world_name, fn_name, raise_if_missing=True):
    """
    Import and return the function from world.

    :param world_module:
        module. a python module encompassing the worlds
    :param world_name:
        string. the name of the world in the module
    :param fn_name:
        string. the name of the function in the world
    :param raise_if_missing:
        bool. if true, raise error if function not found

    :return:
        the function, if defined by the world.
    """
    result_fn = None
    try:
        DesiredWorld = getattr(world_module, world_name)
        result_fn = getattr(DesiredWorld, fn_name)
    except Exception as e:
        if raise_if_missing:
            print("Could not find {} for {}".format(fn_name, world_name))
            raise e
    return result_fn


def get_eligibility_fn(world_module, world_name):
    """
    Get eligibility function for a world.

    :param world_module:
        module. a python module encompassing the worlds
    :param world_name:
        string. the name of the world in the module

    :return:
        the eligibility function if available, else None
    """
    return get_world_fn_attr(
        world_module, world_name, 'eligibility_function', raise_if_missing=False
    )


def get_assign_roles_fn(world_module, world_name):
    """
    Get assign roles function for a world.

    :param world_module:
        module. a python module encompassing the worlds
    :param world_name:
        string. the name of the world in the module

    :return:
        the assign roles function if available, else None
    """
    return get_world_fn_attr(
        world_module, world_name, 'assign_roles', raise_if_missing=False
    )


def default_assign_roles_fn(agents):
    """
    Assign agent role.

    Default role assignment.

    :param:
        list of agents
    """
    for i, a in enumerate(agents):
        a.disp_id = f'Agent_{i}'


class SafetyDetectionResult(Enum):
    """
    Result of identfying offensive language in a response.

    SAFE:       the message is safe
    BLOCKLIST:  the message contains a word from the blocklist
    UNSAFE:     the message is deemed unsafe by the safety classifier
    """

    SAFE = 0
    BLOCKLIST = 1
    UNSAFE = 2


class ReportResult(Enum):
    """
    Result of filing a report.

    FAILURE:    a player timed out while reporting, or it was an accidental report
    BLOCK:      a player is blocked, for having been reported > 1 times
    SUCCESS:    a successful report
    BOT:        the offending agent was the bot
    """

    FAILURE = 0
    BLOCK = 1
    SUCCESS = 2
    BOT = 3


class UploadImageResult(Enum):
    """
    Result of uploading an image.

    SUCCESS:        user successfully uploaded an image
    OBJECTIONABLE:  the image contains objectionable content
    ERROR:          there was an error
    """

    SUCCESS = 0
    OBJECTIONABLE = 1
    ERROR = 2


class PersonalInfoDetector(object):
    """
    Detects whether a string contains any of the following personal information
    datapoints using regular expressions:

    - credit card
    - phone number
    - email
    - SSN
    """

    def __init__(self):
        self.credit_card_regex = r"((?:(?:\\d{4}[- ]?){3}\\d{4}|\\d{15,16}))(?![\\d])"
        self.email_regex = (
            r"([a-z0-9!#$%&'*+\/=?^_`{|.}~-]+@(?:[a-z0-9](?:[a-z0-9-]*"
            + r"[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?)"
        )
        self.phone_number_regex = (
            r"\D?(\d{0,3}?)\D{0,2}(\d{3})?\D{0,2}(\d{3})\D?(\d{4})$"
        )
        self.ssn_regex = r"^\d{3}-\d{2}-\d{4}$"

    def detect_all(self, text):
        contains = {}
        contains["credit_card"] = self.detect_credit_card(text)
        contains["email"] = self.detect_email(text)
        contains["phone_number"] = self.detect_phone_number(text)
        contains["ssn"] = self.detect_ssn(text)
        return contains

    def txt_format_detect_all(self, text):
        contains = self.detect_all(text)
        contains_personal_info = False
        txt = "We believe this text contains the following personal " + "information:"
        for k, v in contains.items():
            if v != []:
                contains_personal_info = True
                txt += f"\n- {k.replace('_', ' ')}: {', '.join([str(x) for x in v])}"
        if not contains_personal_info:
            return ""
        return txt

    def detect_credit_card(self, text):
        return re.findall(self.credit_card_regex, text)

    def detect_email(self, text):
        text = text.lower()
        return re.findall(self.email_regex, text)

    def detect_phone_number(self, text):
        phones = re.findall(self.phone_number_regex, text)
        edited = []
        for tup in phones:
            edited.append("".join(list(tup)))
        return edited

    def detect_ssn(self, text):
        return re.findall(self.ssn_regex, text)


class DictFrequencies:
    """
    Dict freqs.
    """

    def __init__(self, freqs):
        self.freqs = freqs
        self.N = sum(freqs.values())
        self.V = len(freqs)
        self.logNV = math.log(self.N + self.V)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import random
from typing import Any, Dict, List, Optional

from parlai.agents.fixed_response.fixed_response import FixedResponseAgent
from parlai.core.agents import Agent
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.worlds import create_task, DialogPartnerWorld, validate
from parlai.core.message import Message


def load_openers(opt) -> Optional[List[str]]:
    base_task = opt['task'].split(':')[0]
    if base_task == 'self_chat':
        # TODO(#2284): Load default openers from s3
        return None

    print('[ loading conversation openers... ]')
    # create dummy task so we can get openers from the data
    task_opt = copy.deepcopy(opt)
    task_opt['task'] = base_task

    # default train will loop forever, but evalmode will stop after one epoch
    datatype = task_opt['datatype']
    if 'train' in datatype and 'evalmode' not in datatype:
        task_opt['datatype'] = f'{datatype}:evalmode'
    task_opt['interactive_task'] = False
    task_opt['selfchat_task'] = False
    task_opt['fixed_response'] = None
    task_agent = FixedResponseAgent(task_opt)
    task_world = create_task(task_opt, task_agent)

    # run through task data, collecting all first messages
    openers = set()
    is_first_turn = True
    while not task_world.epoch_done():
        task_world.parley()
        msg = task_world.get_acts()[0]
        # add only the first message in the episode
        if is_first_turn and msg.get('text'):
            openers.add(msg['text'])
        is_first_turn = msg.get('episode_done', False)

    print(f'[ loaded {len(openers)} openers ]')
    return list(openers)


class SelfChatWorld(DialogPartnerWorld):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Return the parser as-is.

        Self-chat-specific world flags can be added here.
        """
        return parser

    def __init__(self, opt, agents, shared=None):
        super().__init__(opt, agents, shared)
        self.init_contexts(shared=shared)
        self._openers = None
        self.init_openers()
        self.max_turn_cnt = self.opt.get('selfchat_max_turns', 10)
        self.turn_cnt = 0
        self.episode_cnt = 0

    def init_contexts(self, shared=None) -> None:
        """
        Override to load or instantiate contexts to be used to seed the self chat.
        """
        pass

    def get_contexts(self):
        """
        Override to return a pair of contexts with which to seed the self chat episode.

        This function will be called before the first turn of every episode.
        """
        return ['Hi!', '']

    def init_openers(self) -> None:
        """
        Override to load or instantiate opening messages to be used to seed the self
        chat.
        """
        if self.opt.get('seed_messages_from_task'):
            self._openers = load_openers(self.opt)

    def get_openers(self, episode_num: int) -> Optional[List[str]]:
        """
        Override to return one or more opening messages with which to seed the self chat
        episode.

        The return value should be an array of strings, each string being a message in
        response to the string before it.
        """
        if self._openers:
            return [random.choice(self._openers)]
        return None

    def display(self):
        s = super().display()
        if self.turn_cnt == 0:
            s += '\n==============================\n'
        return s

    def episode_done(self):
        return self.turn_cnt >= self.max_turn_cnt

    def _get_seed_utt_acts(
        self, episode_num: int, agents: List[Agent]
    ) -> List[Dict[str, Any]]:
        def make_agent_action(utterance: str, agent: Agent) -> Dict[str, Any]:
            return {'text': utterance, 'episode_done': False, 'id': agent.id}

        openers = self.get_openers(episode_num)
        if not openers:
            return []
        return list(map(make_agent_action, openers, agents))

    def parley(self):
        if self.episode_done():
            self.turn_cnt = 0
            self.episode_cnt += 1
            self.contexts = None
            self.seed_utterances = None
            agents = self.get_agents()
            for a in agents:
                a.reset()

        if self.turn_cnt == 0:
            self.acts = [None, None]
            # get the beginning of the conversation, which can include contexts
            # and/or any number of starting messages
            self.contexts = self.get_contexts()
            self.seed_utterances = self._get_seed_utt_acts(
                self.episode_cnt, self.agents
            )

        if self.contexts:
            assert len(self.contexts) == 2
            # initial context
            for i in range(0, 2):
                context = Message(
                    {'text': self.contexts[i], 'episode_done': False, 'id': 'context'}
                )
                self.acts[i] = context
                self.agents[i].observe(validate(context))
            # clear contexts so they are only added once per episode
            self.contexts = None
        elif self.seed_utterances:
            # pop the next two seed messages (there may be less or more than 2 total)
            utts = self.seed_utterances[:2]
            self.seed_utterances = self.seed_utterances[2:]
            # process the turn
            for i in [0, 1]:
                # if we have a seed utterance, add it to the conversation
                if len(utts) > i:
                    self.acts[i] = utts[i]
                    if hasattr(self.agents[i], 'self_observe'):
                        self.agents[i].observe({'episode_done': False})
                        self.agents[i].self_observe(self.acts[i])
                else:
                    self.acts[i] = self.agents[i].act()
                self.agents[1 - i].observe(validate(self.acts[i]))
        else:
            # do regular loop
            acts = self.acts
            agents = self.agents
            acts[0] = agents[0].act()
            agents[1].observe(validate(acts[0]))
            acts[1] = agents[1].act()
            agents[0].observe(validate(acts[1]))

        self.update_counters()
        self.turn_cnt += 1



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# Download and build the data if it does not exist.

import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile

RESOURCES = [
    DownloadableFile(
        'https://raw.githubusercontent.com/openai/generating-reviews-discovering-sentiment/master/data/test_binary_sent.csv',
        'test_binary_sent.csv',
        '519cea7ed4d22fe7ec4eccbb3d5ba6d88902a3b15ce129f476aa8364463a9fc7',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/openai/generating-reviews-discovering-sentiment/master/data/train_binary_sent.csv',
        'train_binary_sent.csv',
        '6003623bcb35aad3a446a265b8931b7ccab61fcc10f2e9c1fec916ff67c7be35',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/openai/generating-reviews-discovering-sentiment/master/data/dev_binary_sent.csv',
        'dev_binary_sent.csv',
        'f34c4987fea208fefc2d62a1b42c83a766cbfc7ce58c2a878ef953cf91f01729',
        zipped=False,
    ),
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'SST')
    version = None

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Google The Schema-Guided Dialogue(SGD) Dataset implementation for ParlAI.
"""

import os
import json
from parlai.core.opt import Opt
from parlai.core.teachers import DialogTeacher
from parlai.utils.misc import warn_once
from parlai.core.message import Message
from parlai.core.metrics import AverageMetric, BleuMetric
from parlai.utils.io import PathManager

import parlai.tasks.google_sgd.build as build_


class Text2API2TextTeacher(DialogTeacher):
    """
    Teacher which produces both API calls and NLG responses.
    """

    def __init__(self, opt: Opt, shared=None):
        self.fold = opt['datatype'].split(':')[0]
        opt['datafile'] = self.fold
        self.dpath = os.path.join(opt['datapath'], 'google_sgd')
        if shared is None:
            warn_once(
                "Google SGD is a beta dataset, and format may significantly change."
            )
            build_.build(opt)
        super().__init__(opt, shared)

    def _load_data(self, fold):
        dataset_fold = 'dev' if fold == 'valid' else fold
        fold_path = os.path.join(self.dpath, dataset_fold)
        schema_file = os.path.join(fold_path, 'schema.json')
        with PathManager.open(schema_file, 'r') as f:
            schema_lookup = {}
            for schema in json.load(f):
                schema_lookup[schema['service_name']] = schema

        dialogs = []
        for file_id in range(1, build_.fold_size(dataset_fold) + 1):
            filename = os.path.join(fold_path, f'dialogues_{file_id:03d}.json')
            with PathManager.open(filename, 'r') as f:
                dialogs += json.load(f)
        return schema_lookup, dialogs

    def _get_api_call_and_results(self, sys_turn, schema_lookup):
        api_call = {}
        api_resp = {}
        for frame in sys_turn['frames']:
            if 'service_call' in frame:
                # API CALL
                method = frame['service_call']['method']
                for slot_type, slot_value in frame['service_call'][
                    'parameters'
                ].items():
                    api_call[f'{method}.{slot_type}'] = slot_value
                assert 'service_results' in frame

            # API Resp
            if 'actions' in frame:
                for action in frame['actions']:
                    slot_type = action['slot']
                    slot_value = action['canonical_values']
                    api_resp[slot_type] = slot_value
        return api_call, api_resp

    def custom_evaluation(
        self, teacher_action: Message, labels, model_response: Message
    ):
        resp = model_response.get('text')
        if not resp:
            return

        if teacher_action['type'] == 'apicall' and resp.startswith('apicall: '):
            gold = teacher_action['slots']
            slot_strs = resp[9:].split(' ; ')
            parsed = {}
            for slot_str in slot_strs:
                if ' = ' not in slot_str:
                    if slot_str != '':
                        # syntactically invalid generations should count against us
                        self.metrics.add('slot_p', AverageMetric(0))
                    continue
                name, value = slot_str.split(' = ')
                parsed[name] = value

            # slot precision
            for k, v in parsed.items():
                self.metrics.add('slot_p', AverageMetric(v == gold.get(k)))
            # slot recall
            for k, v in gold.items():
                self.metrics.add('slot_r', AverageMetric(v == parsed.get(k)))
        elif teacher_action['type'] == 'apiresp':
            delex_resp = self._delex(resp, teacher_action['slots'])
            delex_label = self._delex(labels[0], teacher_action['slots'])
            self.metrics.add(
                'delex_bleu', BleuMetric.compute(delex_resp, [delex_label])
            )

    def _delex(self, text, slots):
        delex = text
        for slot, values in slots.items():
            assert isinstance(values, list)
            for value in values:
                delex = delex.replace(value, slot)
        return delex

    def _api_dict_to_str(self, apidict):
        return ' ; '.join(f'{k} = {v}' for k, v in apidict.items())

    def setup_data(self, fold):
        schema_lookup, dialogs = self._load_data(fold)
        for dialog in dialogs:
            # services = dialog['services']
            turns = dialog['turns']
            num_turns = len(turns)
            for turn_id in range(0, num_turns, 2):
                is_first_turn = turn_id == 0

                user_turn = turns[turn_id]
                sys_turn = turns[turn_id + 1]
                api_call, api_results = self._get_api_call_and_results(
                    sys_turn, schema_lookup
                )
                call_str = self._api_dict_to_str(api_call)
                resp_str = self._api_dict_to_str(api_results)
                if not api_call and not api_results:
                    # input: user_turn, output: sys_turn
                    yield {
                        'text': user_turn['utterance'],
                        'label': sys_turn['utterance'],
                        'type': 'text',
                    }, is_first_turn
                elif not api_call and api_results:
                    yield {
                        'text': f"{user_turn['utterance']} api_resp: {resp_str}",
                        'label': sys_turn['utterance'],
                        'type': 'apiresp',
                        'slots': api_results,
                    }, is_first_turn
                elif api_call and api_results:
                    # input: user_turn, output: api_call
                    yield {
                        'text': user_turn['utterance'],
                        'label': f'apicall: {call_str}',
                        'type': 'apicall',
                        'slots': api_call,
                    }, is_first_turn

                    # system turn, input : api results, output : assistant turn
                    yield {
                        'text': f"api_resp: {resp_str}",
                        'label': sys_turn['utterance'],
                        'type': 'apiresp',
                        'slots': api_results,
                    }, False
                else:
                    assert (
                        api_call and api_results
                    ), "API call without API results! Check Dataset!"


class Text2TextTeacher(Text2API2TextTeacher):
    """
    Text-only teacher (with no API calls or slots)
    """

    def setup_data(self, fold):
        schema_lookup, dialogs = self._load_data(fold)
        for dialog in dialogs:
            turns = dialog['turns']
            num_turns = len(turns)
            for turn_id in range(0, num_turns, 2):
                if turn_id == 0:
                    is_first_turn = True
                else:
                    is_first_turn = False

                user_turn = turns[turn_id]
                sys_turn = turns[turn_id + 1]
                # input: user_turn, output: sys_turn
                yield {
                    'text': user_turn['utterance'],
                    'label': sys_turn['utterance'],
                    'type': 'text',
                }, is_first_turn


class DefaultTeacher(Text2API2TextTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
File for Message object and associated functions.

The Message object's key function is to prevent users from editing fields in an action
or observation dict unintentionally.
"""

from __future__ import annotations


class Message(dict):
    """
    Class for observations and actions in ParlAI.

    Functions like a dict, but triggers a RuntimeError when calling __setitem__ for a
    key that already exists in the dict.
    """

    def __setitem__(self, key, val):
        if key in self:
            raise RuntimeError(
                'Message already contains key `{}`. If this was intentional, '
                'please use the function `force_set(key, value)`.'.format(key)
            )
        super().__setitem__(key, val)

    def force_set(self, key, val):
        super().__setitem__(key, val)

    def copy(self):
        return type(self)(self)

    @classmethod
    def padding_example(cls) -> Message:
        """
        Create a Message for batch padding.
        """
        return cls({'batch_padding': True, 'episode_done': True})

    def is_padding(self) -> bool:
        """
        Determine if a message is a padding example or not.
        """
        return bool(self.get('batch_padding'))



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import parlai.core.build_data as build_data
import os
from hashlib import sha1
from parlai.tasks.light_dialog_wild.builder import build_from_dump
from parlai.core.build_data import DownloadableFile

RESOURCES = [
    DownloadableFile(
        'http://parl.ai/downloads/light_project/wild_chats/contents.txt',
        'current_chats.txt',
        'c708fe62692f239a2b35025d71722c7607f863ffa110aa118f2e1d0fa7db4730',
        zipped=False,
    )
]


def download(opt):
    version = 'v0.1'
    # download all of the currently available jsons
    dpath = os.path.join(opt['datapath'], 'light_dialogue_wild', 'dumps')
    if not build_data.built(dpath, version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the manifest.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # Read the manifest
        with open(os.path.join(dpath, 'current_chats.txt'), 'r') as manifest:
            additional_files = manifest.readlines()

        # Download the files listed in the manifest
        more_resources = []
        for sha_and_filename in additional_files:
            file_hash, filename = sha_and_filename.strip().split('  ')
            more_resources.append(
                DownloadableFile(
                    f'http://parl.ai/downloads/light_project/wild_chats/{filename}',
                    filename,
                    file_hash,
                    zipped=False,
                )
            )

        for downloadable_file in more_resources:
            downloadable_file.download_file(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version)

    return dpath, version


def get_fpath(opt):
    fields = [
        'taskname',
        'setting',
        'person_names',
        'persona',
        'emote',
        'speech',
        'action',
        'affordances',
        'repeat',
        'cands',
        'current_self_output',
        'clip_cands',
        'speech_prefix',
        'person_names_prefix',
        'score_cutoff',
        'hard_score_cutoff',
        'max_score_cutoff',
        'date_cutoff',
        'continue_type',
        'unseen_test',
    ]
    fpath = ''
    for f in fields:
        fpath += f + str(opt['light_use_' + f]) + "_"
    if opt.get('light_model_name'):
        fpath += f"model_name{opt['light_model_name']}_"
    return str(sha1(fpath[:-1].encode('utf-8')).hexdigest())


def build(opt):
    dpath, version = download(opt)
    if 'light_use_speech_prefix' not in opt:
        opt['light_use_speech_prefix'] = True
    # create particular instance of dataset depending on flags..
    fpath = get_fpath(opt)
    dump_path = os.path.join(opt['datapath'], 'light_dialogue_wild', 'dumps')
    data_path = os.path.join(opt['datapath'], 'light_dialogue_wild', fpath)
    if not build_data.built(data_path, version):
        if build_data.built(data_path):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(data_path)
        build_data.make_dir(data_path)
        build_from_dump(opt, data_path, dump_path)
        # Mark the data as built.
        build_data.mark_done(data_path, version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from parlai.core.teachers import FixedDialogTeacher
from parlai.tasks.md_gender.build import build  # noqa: F401


class OpensubtitlesTeacher(FixedDialogTeacher):
    def __init__(self, opt, shared):
        raise RuntimeError('Data coming soon!')



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# Download and build the data if it does not exist.

from parlai.core.teachers import DialogTeacher
from parlai.utils.io import PathManager
from .build import build
import os
import json
import copy

FILE_START = 'woz_'
FILE_END = '_en.json'


def _path(opt):
    build(opt)

    dt = opt['datatype'].split(':')[0]

    if dt == 'train':
        suffix = 'train'
    # Using matched set as valid and mismatched set as test
    elif dt == 'valid':
        suffix = 'validate'
    elif dt == 'test':
        suffix = 'test'
    else:
        raise RuntimeError('Not valid datatype.')

    data_path = os.path.join(opt['datapath'], 'WoZ', FILE_START + suffix + FILE_END)
    return data_path


class WoZTeacher(DialogTeacher):
    def __init__(self, opt, shared=None):

        opt = copy.deepcopy(opt)
        data_path = _path(opt)

        opt['datafile'] = data_path

        # store datatype
        self.dt = opt['datatype'].split(':')[0]

        # store identifier for the teacher in the dialog
        self.id = 'woz'

        build(opt)

        super().__init__(opt, shared)

    def setup_data(self, input_path):
        print('loading: ' + input_path)

        new_episode = True

        with PathManager.open(input_path) as file:
            data = json.load(file)

        for dialogue in data:
            for line in dialogue['dialogue']:
                answer = [':'.join(turn_label) for turn_label in line['turn_label']]
                question = "What is the change in the dialogue state?"
                context = line['transcript']
                if answer:
                    yield (context + '\n' + question, answer, None, None), new_episode


class DefaultTeacher(WoZTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Taskmaster-2 implementation for ParlAI.

No official train/valid/test splits are available as of 2020-05-18, so we make our own
splits.
"""

from parlai.core.params import ParlaiParser
import os
import pandas as pd
import hashlib
from collections import Counter
from parlai.core.opt import Opt
from parlai.core.teachers import DialogTeacher
from parlai.core.metrics import AverageMetric, F1Metric, BleuMetric
from parlai.utils.misc import warn_once
import json
import parlai.utils.logging as logging
from typing import Optional, Tuple
from parlai.core.message import Message
from parlai.utils.io import PathManager

import parlai.tasks.taskmaster2.build as build_

DOMAINS = [
    'flights',
    'food-ordering',
    'hotels',
    'movies',
    'restaurant-search',
    'sports',
    'music',
]

ONTO_TOKEN = "Onto:"
CALL_TOKEN = "Call:"
RESP_TOKEN = "Result:"


class _Abstract(DialogTeacher):
    """
    Abstract data loader.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        parser.add_argument('--include-ontology', type=bool, default=False)
        parser.add_argument(
            '--domains',
            nargs='+',
            default=DOMAINS,
            choices=DOMAINS,
            help='Uses last passed in configuration.',
        )
        return parser

    def __init__(self, opt: Opt, shared=None):
        self.fold = opt['datatype'].split(':')[0]
        opt['datafile'] = self.fold
        self.dpath = os.path.join(opt['datapath'], 'taskmaster-2')
        if shared is None:
            warn_once(
                "Taskmaster2 is a beta dataset, and format may significantly change."
            )
            build_.build(opt)
        super().__init__(opt, shared)

    def _h(self, x):
        """
        Hash function.
        """
        h = int(hashlib.sha1(x.encode('utf-8')).hexdigest(), 16) % 10
        if h == 0:
            return 'valid'
        elif h == 1:
            return 'test'
        else:
            return 'train'

    def _normalize_annotation(self, anno):
        return anno

    def _load_data(self, fold, domains):
        # load up the ontology
        ontology = {}
        for section in domains:
            parts = []
            fn = os.path.join(self.dpath, section + '.onto.json')
            with PathManager.open(fn, 'r') as f:
                o = json.load(f)
            assert len(o) == 1
            o = list(o.values())[0]
            for sub in o:
                prefix = sub['prefix']
                parts += [
                    self._normalize_annotation(f'{prefix}.{a}')
                    for a in sub['annotations']
                ]
            ontology[section] = ' ; '.join(parts)

        chunks = []
        for section in domains:
            with PathManager.open(os.path.join(self.dpath, section + '.json')) as f:
                subset = pd.read_json(f)
            subset['domain'] = section
            chunks.append(subset)
        chunks = pd.concat(chunks, axis=0)
        # shuffle deterministically for randomness in few-shot training
        chunks = chunks.sample(frac=1.0, random_state=42)
        chunks['fold'] = self._label_fold(chunks)
        # only the fold we need here
        chunks = chunks[chunks.fold == fold].reset_index()
        chunks['ontology'] = chunks['domain'].apply(ontology.get)
        return chunks

    def _segments2text(self, segments):
        output = []
        slots = {}
        for segment in segments:
            val = segment['text']
            for anno_ in segment['annotations']:
                anno = anno_['name']
                anno = self._normalize_annotation(anno)
                output.append(f'{anno} = {val}')
                slots[anno] = val
        return " ; ".join(output), slots

    def custom_evaluation(
        self,
        teacher_action: Message,
        labels: Optional[Tuple[str]],
        model_response: Message,
    ):
        if 'metrics' in model_response and 'type' in teacher_action:
            # keep copies of metrics across both api calls/responses
            prefix = teacher_action['type']
            keys = list(model_response['metrics'].keys())
            for k in keys:
                self.metrics.add(f'{prefix}_{k}', model_response['metrics'][k])

        if 'text' not in model_response or not labels or 'type' not in teacher_action:
            return

        domain = teacher_action['domain']

        if teacher_action['type'] == 'apicall':
            # also count slot accuracy
            text = model_response['text']
            slot_guesses = set(
                text.replace(CALL_TOKEN + " ", "").split(' ; ')
            )  # prevent cheating via repeated guesses
            correct = 0
            for slot_guess in slot_guesses:
                if ' = ' not in slot_guess:
                    continue
                try:
                    slot, guess = slot_guess.split(' = ')
                except ValueError:
                    continue
                if teacher_action['slots'].get(slot) == guess:
                    self.metrics.add('slot_p', AverageMetric(1))
                    self.metrics.add(f'{domain}_slot_p', AverageMetric(1))
                    correct += 1
                else:
                    self.metrics.add('slot_p', AverageMetric(0))
                    self.metrics.add(f'{domain}_slot_p', AverageMetric(0))
                    logging.debug(
                        f"Bad slot guess '{slot_guess}' != {teacher_action['slots']}"
                    )
            if teacher_action['slots']:
                self.metrics.add(
                    'slot_r', AverageMetric(correct, len(teacher_action['slots']))
                )
                self.metrics.add(
                    f'{domain}_slot_r',
                    AverageMetric(correct, len(teacher_action['slots'])),
                )
                self.metrics.add(
                    'jga', AverageMetric(correct == len(teacher_action['slots']))
                )

        elif teacher_action['type'] == 'apiresp':
            # keep track of statistics by domain
            f1_metric = F1Metric.compute(model_response['text'], labels)
            bleu_metric = BleuMetric.compute(model_response['text'], labels)
            self.metrics.add(f'{domain}_lex_f1', f1_metric)
            self.metrics.add(f'{domain}_lex_bleu', bleu_metric)

            delex_text = model_response['text']
            delex_label = labels[0]
            # compute delexicalized string metrics
            for slot, value in teacher_action['slots'].items():
                delex_text = delex_text.replace(value, slot)
                delex_label = delex_label.replace(value, slot)
            f1_metric = F1Metric.compute(delex_text, (delex_label,))
            self.metrics.add('delex_f1', f1_metric)
            self.metrics.add(f'{domain}_delex_f1', f1_metric)
            bleu_metric = BleuMetric.compute(delex_text, [delex_label])
            self.metrics.add('delex_bleu', bleu_metric)
            self.metrics.add(f'{domain}_delex_bleu', bleu_metric)

    def setup_data(self, fold):
        domains = self.opt.get('domains', DOMAINS)
        chunks = self._load_data(fold, domains)
        domains_cnt = Counter()
        for _, row in chunks.iterrows():
            domains_cnt[row['domain']] += 1
            first = True
            utterances = row['utterances'][:]
            if (
                len(utterances) >= 3
                and utterances[0]['speaker'] == 'USER'
                and utterances[1]['speaker'] == 'ASSISTANT'
                and utterances[2]['speaker'] == 'ASSISTANT'
                and "help you?" in utterances[1]['text']
            ):
                # skip this one
                utterances.pop(1)
            if self.opt['include_ontology']:
                yield {'text': f"{ONTO_TOKEN} {row['ontology']}", 'label': ''}, True
                first = False
            while utterances:
                utt = utterances.pop(0)
                segtxt, slots = self._segments2text(utt.get('segments', []))
                if utt['speaker'] == 'USER':
                    yield {
                        'text': utt['text'],
                        'label': f'{CALL_TOKEN} {segtxt}',
                        'domain': row['domain'],
                        'slots': slots,
                        'type': 'apicall',
                    }, first
                    first = False
                elif utt['speaker'] == 'ASSISTANT':
                    yield {
                        'text': f'{RESP_TOKEN} {segtxt}',
                        'label': utt['text'],
                        'domain': row['domain'],
                        'slots': slots,
                        'type': 'apiresp',
                    }, first
                    first = False
        logging.debug(f"Fold {fold} domains: {domains_cnt}")


class DelexTeacher(_Abstract):
    def _label_fold(self, chunks):
        return chunks.conversation_id.apply(self._h)

    def _delexicalize(self, text, slots):
        for key, value in slots.items():
            text = text.replace(value, key)
        return text

    def setup_data(self, fold):
        domains_cnt = Counter()
        chunks = self._load_data(fold)
        for _, row in chunks.iterrows():
            domains_cnt[row['domain']] += 1
            first = True
            utterances = row['utterances'][:]
            if (
                len(utterances) >= 3
                and utterances[0]['speaker'] == 'USER'
                and utterances[1]['speaker'] == 'ASSISTANT'
                and utterances[2]['speaker'] == 'ASSISTANT'
                and "help you?" in utterances[1]['text']
            ):
                # skip this one
                utterances.pop(1)

            user_utterances = []
            asst_utterances = []
            while utterances:
                utt = utterances.pop(0)
                _, slots = self._segments2text(utt.get('segments', []))
                if utt['speaker'] == 'USER':
                    if asst_utterances:
                        yield {
                            'text': ' __BREAK__ '.join(user_utterances),
                            'label': ' __BREAK__ '.join(asst_utterances),
                            'domain': row['domain'],
                        }, first
                        first = False
                        user_utterances = []
                        asst_utterances = []
                    user_utterances.append(self._delexicalize(utt['text'], slots))
                elif utt['speaker'] == 'ASSISTANT':
                    asst_utterances.append(self._delexicalize(utt['text'], slots))
                    if not user_utterances:
                        user_utterances.append('__SILENCE__')
            if asst_utterances:
                yield {
                    'text': ' __BREAK__ '.join(user_utterances),
                    'label': ' __BREAK__ '.join(asst_utterances),
                    'domain': row['domain'],
                }, first


class TextOnlyTeacher(DelexTeacher):
    def _delexicalize(self, text, slots):
        return text


class FullShotTeacher(_Abstract):
    """
    The full shot teacher uses a standard 80-10-10 split, without regarding domain.
    """

    def _label_fold(self, chunks):
        return chunks.conversation_id.apply(self._h)


class FewShotTeacher(_Abstract):
    """
    Few shot teacher tests for generalization to new domains.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        parser.add_argument(
            '--holdout',
            default=DOMAINS[0],
            choices=DOMAINS,
            help='Domain which is held out from test',
        )
        parser.add_argument(
            '--n-shot',
            default=100,
            type=int,
            help='Number of few shot examples to provide in training fold.',
        )
        return super().add_cmdline_args(parser, partial_opt=partial_opt)

    def _label_fold(self, chunks):
        folds = []
        num_shots = 0
        for _, row in chunks.iterrows():
            if row['domain'] != self.opt['holdout']:
                # if it's not in the holdout, always mark it train
                folds.append('train')
            else:
                # keep the same valid/test sets as in fullshot, and only leak
                # a small number of the training examples (i.e. throw away the
                # vast majority of our data but keep test sets the same)

                f = self._h(row['conversation_id'])
                if f != 'train':
                    folds.append(f)
                elif num_shots < self.opt['n_shot']:
                    folds.append('train')
                    num_shots += 1
                else:
                    folds.append('throwaway')
        return folds


class DefaultTeacher(FullShotTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from functools import lru_cache

import torch

from parlai.core.torch_ranker_agent import TorchRankerAgent

from .modules import MemNN, opt_to_kwargs


class MemnnAgent(TorchRankerAgent):
    """
    Memory Network agent.

    Tips:
    - time features are necessary when memory order matters
    - multiple hops allow multiple steps of reasoning, but also seem to make it
        easier to learn to read the memories if you have at least two hops
    - 'adam' seems to work very poorly compared to 'sgd' for hogwild training
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        arg_group = parser.add_argument_group('MemNN Arguments')
        arg_group.add_argument(
            '-esz',
            '--embedding-size',
            type=int,
            default=128,
            help='size of token embeddings',
        )
        arg_group.add_argument(
            '-hops', '--hops', type=int, default=3, help='number of memory hops'
        )
        arg_group.add_argument(
            '--memsize',
            type=int,
            default=32,
            help='size of memory, set to 0 for "nomemnn" model which just '
            'embeds query and candidates and picks most similar candidate',
        )
        arg_group.add_argument(
            '-tf',
            '--time-features',
            type='bool',
            default=True,
            help='use time features for memory embeddings',
        )
        arg_group.add_argument(
            '-pe',
            '--position-encoding',
            type='bool',
            default=False,
            help='use position encoding instead of bag of words embedding',
        )
        parser.set_defaults(
            split_lines=True, add_p1_after_newln=True, encode_candidate_vecs=True
        )
        super().add_cmdline_args(parser, partial_opt=partial_opt)
        cls.dictionary_class().add_cmdline_args(parser, partial_opt=partial_opt)
        return arg_group

    def __init__(self, opt, shared=None):
        self.id = 'MemNN'
        self.memsize = opt['memsize']
        if self.memsize < 0:
            self.memsize = 0
        self.use_time_features = opt['time_features']
        super().__init__(opt, shared)

    def build_dictionary(self):
        """
        Add the time features to the dictionary before building the model.
        """
        d = super().build_dictionary()
        if self.use_time_features:
            # add time features to dictionary before building the model
            for i in range(self.memsize):
                d[self._time_feature(i)] = 100_000_000 + i
        return d

    def build_model(self):
        """
        Build MemNN model.
        """
        kwargs = opt_to_kwargs(self.opt)
        return MemNN(
            len(self.dict),
            self.opt['embedding_size'],
            padding_idx=self.NULL_IDX,
            **kwargs,
        )

    def _score(self, output, cands):
        if cands.dim() == 2:
            return torch.matmul(output, cands.t())
        elif cands.dim() == 3:
            return torch.bmm(output.unsqueeze(1), cands.transpose(1, 2)).squeeze(1)
        else:
            raise RuntimeError(
                'Unexpected candidate dimensions {}' ''.format(cands.dim())
            )

    def encode_candidates(self, padded_cands):
        return self.model.answer_embedder(padded_cands)

    def score_candidates(self, batch, cand_vecs, cand_encs=None):
        mems = self._build_mems(batch.memory_vecs)
        # Check for rows that have no non-null tokens
        pad_mask = None
        if mems is not None:
            pad_mask = (mems != self.NULL_IDX).sum(dim=-1) == 0

        if cand_encs is not None:
            state, _ = self.model(batch.text_vec, mems, None, pad_mask)
        else:
            state, cand_encs = self.model(batch.text_vec, mems, cand_vecs, pad_mask)
        scores = self._score(state, cand_encs)

        return scores

    @lru_cache(maxsize=None)  # bounded by opt['memsize'], cache string concats
    def _time_feature(self, i):
        """
        Return time feature token at specified index.
        """
        return '__tf{}__'.format(i)

    def vectorize(self, *args, **kwargs):
        """
        Override options in vectorize from parent.
        """
        kwargs['add_start'] = False
        kwargs['add_end'] = False
        return super().vectorize(*args, **kwargs)

    def batchify(self, obs_batch, sort=False):
        """
        Override so that we can add memories to the Batch object.
        """
        batch = super().batchify(obs_batch, sort)

        # get valid observations
        valid_obs = [(i, ex) for i, ex in enumerate(obs_batch) if self.is_valid(ex)]

        if len(valid_obs) == 0:
            return batch

        valid_inds, exs = zip(*valid_obs)

        # get memories for the valid observations
        mems = None
        if any('memory_vecs' in ex for ex in exs):
            mems = [ex.get('memory_vecs', None) for ex in exs]
        batch.memory_vecs = mems
        return batch

    def _set_text_vec(self, obs, history, truncate):
        """
        Override from Torch Agent so that we can use memories.
        """
        if 'text' not in obs:
            return obs

        if 'text_vec' not in obs:
            # text vec is not precomputed, so we set it using the history
            obs['full_text'] = history.get_history_str()
            history_vecs = history.get_history_vec_list()
            if len(history_vecs) > 0:
                obs['memory_vecs'] = history_vecs[:-1]
                obs['text_vec'] = history_vecs[-1]
            else:
                obs['memory_vecs'] = []
                obs['text_vec'] = []

        # check truncation
        if 'text_vec' in obs:
            truncated_vec = self._check_truncate(obs['text_vec'], truncate, True)
            obs.force_set('text_vec', torch.LongTensor(truncated_vec))

        if 'memory_vecs' in obs:
            obs.force_set(
                'memory_vecs',
                [
                    torch.LongTensor(self._check_truncate(m, truncate, True))
                    for m in obs['memory_vecs']
                ],
            )

        return obs

    def _build_mems(self, mems):
        """
        Build memory tensors.

        During building, will add time features to the memories if enabled.

        :param mems:
            list of length batchsize containing inner lists of 1D tensors
            containing the individual memories for each row in the batch.

        :returns:
            3d padded tensor of memories (bsz x num_mems x seqlen)
        """
        if mems is None:
            return None
        bsz = len(mems)
        if bsz == 0:
            return None

        num_mems = max(len(mem) for mem in mems)
        if num_mems == 0 or self.memsize <= 0:
            return None
        elif num_mems > self.memsize:
            # truncate to memsize most recent memories
            num_mems = self.memsize
            mems = [mem[-self.memsize :] for mem in mems]

        try:
            seqlen = max(len(m) for mem in mems for m in mem)
            if self.use_time_features:
                seqlen += 1  # add time token to each sequence
        except ValueError:
            return None

        padded = torch.LongTensor(bsz, num_mems, seqlen).fill_(0)

        for i, mem in enumerate(mems):
            tf_offset = len(mem) - 1
            for j, m in enumerate(mem):
                padded[i, j, : len(m)] = m
                if self.use_time_features:
                    padded[i, j, -1] = self.dict[self._time_feature(tf_offset - j)]

        if self.use_cuda:
            padded = padded.cuda()

        return padded



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.agents.hugging_face.dict import DialoGPTDictionaryAgent
from parlai.agents.hugging_face.gpt2 import Gpt2Agent, GPT2Decoder, HFGPT2Model
from parlai.utils.misc import warn_once

try:
    from transformers import GPT2Model
except ImportError:
    raise ImportError('Please run `pip install transformers`.')


############################################
## Modules
############################################


class DialoGPTDecoder(GPT2Decoder):
    """
    DialoGPT Decoder.

    This decoder is initialized with the pretrained model from Hugging Face.
    """

    def __init__(self, opt, dict):
        super().__init__(opt, dict)
        self.NULL_IDX, self.START_IDX, self.END_IDX = self._get_special_tokens(
            opt, dict
        )

    @staticmethod
    def _get_special_tokens(opt, dict):
        null_idx = dict.null_idx
        if (
            opt.get('batchsize', 1) == 1
            and not opt['add_special_tokens']
            and null_idx == dict.end_idx
        ):
            # get around the dual usage of end_idx that would otherwise mask endtoken during forward pass.
            null_idx = -1
        return null_idx, dict.start_idx, dict.end_idx

    def _init_from_pretrained(self, opt):
        # load model
        model_sz = opt['gpt2_size']
        fle_key = f'microsoft/DialoGPT-{model_sz}'
        return GPT2Model.from_pretrained(fle_key)


class DialoGPTModel(HFGPT2Model):
    """
    Hugging Face DialoGPT Model.
    """

    def _get_special_tokens(self, opt, dict):
        # keep it consistent between DialoGPTModel and DialoGPTDecoder on start_idx, end_idx, null_idx
        return DialoGPTDecoder._get_special_tokens(opt, dict)

    def _get_decoder(self, opt, dict):
        return DialoGPTDecoder(opt, dict)


############################################
## Agent
############################################


class DialogptAgent(Gpt2Agent):
    """
    Hugging Face DialoGPT Agent.

    DialoGPT is based on GPT2, which is a multi-layer decoder-only Transformer.
    The decoder is initialized with pretrained weights from Hugging Face.
    Read more about this model here
    <https://huggingface.co/transformers/model_doc/dialogpt.html>.

    DialoGPT comes in three  sizes: small, medium, large.

    If you are finetuning the Dialogpt Agent as a dialogue agent, be sure
    to run `--add-special-tokens True`. To examine the performance of the
    agent out of the box, run with `--add-special-tokens False`, and make
    sure that the batch size is 1.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        agent = parser.add_argument_group('DialoGPT Args')
        agent.add_argument(
            '--gpt2-size',
            type=str,
            default='medium',
            choices=['small', 'medium', 'large'],
            help='Which size model to initialize.',
        )
        parser.set_defaults(
            delimiter='<|endoftext|>',
            history_add_global_end_token='<|endoftext|>',
            text_truncate=768,
            label_truncate=256,
            dict_maxexs=0,  # skip building dictionary
        )
        super().add_cmdline_args(parser, partial_opt=partial_opt)
        warn_once('WARNING: this model is in beta and the API is subject to change.')
        return agent

    @staticmethod
    def dictionary_class():
        """
        Return the dictionary class that this agent expects to use.

        Can be overriden if a more complex dictionary is required.
        """
        return DialoGPTDictionaryAgent

    def build_model(self, states=None):
        """
        Build and return model.
        """
        return DialoGPTModel(self.opt, self.dict)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Documents, in a sqlite database.

Adapted from Adam Fisch's work at github.com/facebookresearch/DrQA/
"""

import sqlite3
from . import utils


class DocDB(object):
    """
    Sqlite backed document storage.

    Implements get_doc_text(doc_id).
    """

    def __init__(self, db_path=None):
        self.path = db_path
        self.connection = sqlite3.connect(self.path, check_same_thread=False)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        """
        Close the connection to the database.
        """
        self.connection.close()

    def get_doc_ids(self):
        """
        Fetch all ids of docs stored in the db.
        """
        cursor = self.connection.cursor()
        cursor.execute("SELECT id FROM documents")
        results = [r[0] for r in cursor.fetchall()]
        cursor.close()
        return results

    def get_doc_text(self, doc_id):
        """
        Fetch the raw text of the doc for 'doc_id'.
        """
        cursor = self.connection.cursor()
        cursor.execute(
            "SELECT text FROM documents WHERE id = ?", (utils.normalize(doc_id),)
        )
        result = cursor.fetchone()
        cursor.close()
        return result if result is None else result[0]

    def get_doc_value(self, doc_id):
        """
        Fetch the raw text of the doc for 'doc_id'.
        """
        cursor = self.connection.cursor()
        cursor.execute(
            "SELECT value FROM documents WHERE id = ?", (utils.normalize(doc_id),)
        )
        result = cursor.fetchone()
        cursor.close()
        return result if result is None else result[0]

    def add(self, triples):
        cursor = self.connection.cursor()
        cursor.executemany('INSERT OR IGNORE INTO documents VALUES (?,?,?)', triples)
        cursor.close()
        self.connection.commit()
